import logging

from scipy.sparse import triu

from .matrixFile import MatrixFile

log = logging.getLogger(__name__)

class Ginteractions(MatrixFile):

    def __init__(self, pMatrixFile):
        super().__init__(pMatrixFile)

    def load(self):
        log.error('Not implemented')

    def save(self, pFileName, pSymmetric=None, pApplyCorrection=None):

        # self.restoreMaskedBins()
        log.debug(self.matrix.shape)
        mat_coo = triu(self.matrix, k=0, format='csr').tocoo()
        with open(f"{pFileName}.tsv", 'w', encoding='utf-8') as fileh:
            for idx, counts in enumerate(mat_coo.data):
                chr_row, start_row, end_row, _ = self.cut_intervals[mat_coo.row[idx]]
                chr_col, start_col, end_col, _ = self.cut_intervals[mat_coo.col[idx]]
                fileh.write(f"{chr_row}\t{int(start_row)}\t{int(end_row)}\t{chr_col}\t{int(start_col)}\t{int(end_col)}\t{counts}\n")



import logging
import os

import cooler
import numpy as np

from .matrixFile import MatrixFile

log = logging.getLogger(__name__)

class Scool(MatrixFile):

    def __init__(self, pMatrixFile=None):
        super().__init__(pMatrixFile)
        log.debug('scool object created')
        self.coolObjectsList = None
        self.bins = None
        self.pixel_list = None
        self.name_list = None

    def load(self):
        raise NotImplementedError('Please use the specific cell to load the individual cool file from the scool file')

    def save(self, pFileName, pSymmetric=True, pApplyCorrection=True):

        pixel_dict = {}
        bins_dict = {}

        if self.coolObjectsList is not None:
            for coolObject in self.coolObjectsList:
                bins_data_frame, matrix_data_frame, dtype_pixel, _ = coolObject.matrixFile.create_cooler_input(pSymmetric=pSymmetric, pApplyCorrection=pApplyCorrection)
                bins_dict[coolObject.matrixFile.matrixFileName] = bins_data_frame
                pixel_dict[coolObject.matrixFile.matrixFileName] = matrix_data_frame

        else:
            try:
                dtype_pixel = {'bin1_id': np.int32, 'bin2_id': np.int32, 'count': self.pixel_list[0]['count'].dtype}
                # dtype_pixel = self.pixel_list[0]['count'].dtype

                for i, pixels in enumerate(self.pixel_list):
                    bins_dict[self.name_list[i]] = self.bins
                    pixel_dict[self.name_list[i]] = pixels
                    log.debug('self.name_list[i] %s', self.name_list[i])
            except Exception as exp:  # pylint: disable=W0718
                log.debug('Exception %s', str(exp))

        local_temp_dir = os.path.dirname(os.path.realpath(pFileName))

        cooler.create_scool(cool_uri=pFileName, bins=bins_dict, cell_name_pixels_dict=pixel_dict,
                            dtypes=dtype_pixel,
                            ordered=True,
                            temp_dir=local_temp_dir)



import importlib
import logging

log = logging.getLogger(__name__)


class MatrixFileHandler():
    """
    This class handles the load and save of the different Hi-C contact matrix formats.
    """

    def __init__(self, pFileType='cool', pMatrixFile=None, pChrnameList=None,
                 pApplyCorrectionCoolerLoad=None, pBedFileHicPro=None, pCorrectionFactorTable=None,
                 pCorrectionOperator=None, pEnforceInteger=None, pAppend=None, pFileWasH5=None, pHiCInfo=None, pHic2CoolVersion=None,
                 pDistance=None, pMatrixFormat=None, pLoadMatrixOnly=None, pNoCutIntervals=None):

        self.class_ = getattr(importlib.import_module('.' + pFileType.lower(), package='hicmatrix.lib'), pFileType.title())

        if pFileType == 'hicpro':
            self.matrixFile = self.class_(pMatrixFile=pMatrixFile, pBedFile=pBedFileHicPro)
        else:
            self.matrixFile = self.class_(pMatrixFile=pMatrixFile)
            if pFileType == 'cool':
                self.matrixFile.chrnameList = pChrnameList
                if pCorrectionFactorTable is not None:
                    self.matrixFile.correctionFactorTable = pCorrectionFactorTable
                if pCorrectionOperator is not None:
                    self.matrixFile.correctionOperator = pCorrectionOperator
                if pEnforceInteger is not None:
                    self.matrixFile.enforceInteger = pEnforceInteger
                if pAppend is not None:
                    self.matrixFile.appendData = pAppend
                if pFileWasH5 is not None:
                    self.matrixFile.fileWasH5 = pFileWasH5
                if pApplyCorrectionCoolerLoad is not None:
                    self.matrixFile.applyCorrectionLoad = pApplyCorrectionCoolerLoad
                if pHiCInfo is not None:
                    self.matrixFile.hic_metadata = pHiCInfo
                if pHic2CoolVersion is not None:
                    self.matrixFile.hic2cool_version = pHic2CoolVersion
                if pDistance is not None:
                    self.matrixFile.distance = pDistance
                if pMatrixFormat is not None:
                    self.matrixFile.matrixFormat = pMatrixFormat
                if pLoadMatrixOnly is not None:
                    self.matrixFile.matrixOnly = pLoadMatrixOnly
                if pNoCutIntervals is not None:
                    self.matrixFile.noCutIntervals = pNoCutIntervals

    def load(self):

        return self.matrixFile.load()

    def set_matrix_variables(self, pMatrix, pCutIntervals, pNanBins, pCorrectionFactors, pDistanceCounts):
        self.matrixFile.set_matrix_variables(pMatrix, pCutIntervals, pNanBins, pCorrectionFactors, pDistanceCounts)

    def save(self, pName, pSymmetric, pApplyCorrection):
        self.matrixFile.save(pName, pSymmetric, pApplyCorrection)

    def load_init(self):
        pass



import logging
from collections import Counter
from collections import OrderedDict

import time

import cooler
import numpy as np
from intervaltree import Interval, IntervalTree
from scipy.sparse import csr_matrix, dia_matrix
from scipy.sparse import hstack as sparse_hstack
from scipy.sparse import tril, triu
from scipy.sparse import vstack as sparse_vstack
from scipy.sparse import diags
from scipy.sparse import lil_matrix

from .lib import MatrixFileHandler
from .utilities import check_chrom_str_bytes, toBytes, toString

log = logging.getLogger(__name__)

class hiCMatrix:
    """
    Class to handle Hi-C matrices
    contains routines to get intrachromosomal distances
    get sub matrices by chrname.
    """

    def __init__(self, pMatrixFile=None, pChrnameList=None, pDistance=None, pNoIntervalTree=None, pUpperTriangleOnly=None,
                 pMatrixFormat=None, pRestoreMaskedBins=None, pLoadMatrixOnly=None):
        self.non_homogeneous_warning_already_printed = False
        self.bin_size = None
        self.bin_size_homogeneous = None  # track if the bins are equally spaced or not
        self.uncorrected_matrix = None

        self.matrix = None
        self.cut_intervals = None
        self.nan_bins = None
        self.correction_factors = None
        self.distance_counts = None
        # # when NaN bins are masked, this variable becomes contains the bin index
        # # needed to put the masked bins back into the matrix.
        self.orig_bin_ids = []
        self.orig_cut_intervals = []  # similar to orig_bin_ids. Used to identify the position of masked nan bins
        self.matrixFileHandler = None
        start_time = time.time()
        if pMatrixFile is not None:
            log.debug('Load self.matrixFileHandler')
            fileType = 'cool'
            if pMatrixFile.endswith('.h5'):
                fileType = 'h5'
            self.matrixFileHandler = MatrixFileHandler(pFileType=fileType, pMatrixFile=pMatrixFile, pChrnameList=pChrnameList, pDistance=pDistance, pMatrixFormat=pMatrixFormat, pLoadMatrixOnly=pLoadMatrixOnly)
            log.debug('init time: %s', time.time() - start_time)
            matrixFileHandler_load = self.matrixFileHandler.load()
            # check if there was any exception thrown in the load function
            if len(matrixFileHandler_load) == 2:
                raise ValueError(f'Matrix failed to load: {matrixFileHandler_load[1]}')
            self.matrix, self.cut_intervals, self.nan_bins, \
                self.correction_factors, self.distance_counts = matrixFileHandler_load
            if pLoadMatrixOnly is None or not pLoadMatrixOnly:
                if self.nan_bins is None:
                    self.nan_bins = np.array([])

                if pUpperTriangleOnly is None or not pUpperTriangleOnly:
                    self.fillLowerTriangle()
                start_time = time.time()

                if pRestoreMaskedBins is None or pRestoreMaskedBins:
                    self.restoreMaskedBins()
                start_time = time.time()

                if pNoIntervalTree is None or not pNoIntervalTree:
                    self.interval_trees, self.chrBinBoundaries = \
                        self.intervalListToIntervalTree(self.cut_intervals)
                else:
                    log.debug('no intervaltree')

        elif pMatrixFile is None:
            log.debug('Only init object, no matrix given.')
        else:
            raise ValueError('matrix file not given')
        log.debug('data loaded!')

    def save(self, pMatrixName, pSymmetric=True, pApplyCorrection=False, pHiCInfo=None):
        """ As an output format cooler and mcooler are supported.
        """

        if self.matrixFileHandler is None:
            fileType = 'cool'
            if pMatrixName.endswith('h5'):
                fileType = 'h5'
            self.matrixFileHandler = MatrixFileHandler(pFileType=fileType, pHiCInfo=pHiCInfo)

        self.restoreMaskedBins()
        self.matrixFileHandler.set_matrix_variables(self.matrix, self.cut_intervals, self.nan_bins,
                                                    self.correction_factors, self.distance_counts)
        if pMatrixName.endswith('cool'):
            self.matrixFileHandler.matrixFile.hic_metadata = pHiCInfo

        if pMatrixName.endswith('cool') or pMatrixName.endswith('h5'):
            self.matrixFileHandler.save(pMatrixName, pSymmetric=pSymmetric, pApplyCorrection=pApplyCorrection)

    def getInformationCoolerBinNames(self):
        log.info('The following columns are available: %s', self.matrixFileHandler.matrixFile.getInformationCoolerBinNames())

    def fillLowerTriangle(self):
        """
        checks if the matrix is complete or if only half of the matrix was saved.
        Returns a whole matrix.
        """
        # log.debug('sum of tril: {}'.format(tril(self.matrix, k=-1).sum()))
        if tril(self.matrix, k=-1).sum() == 0:
            # this case means that the lower triangle of the
            # symmetric matrix (below the main diagonal)
            # is zero. In this case, replace the lower
            # triangle using the upper triangle
            self.matrix = self.matrix + triu(self.matrix, 1).T

        # return matrix

    def setCutIntervals(self, cut_intervals):
        """
        Replace the cut_intervals of a matrix
        """

        # check that the matrix is squared
        if len(cut_intervals) != self.matrix.shape[0]:
            raise ValueError(f"Length of cut_intervals {len(cut_intervals)} does not match the "
                             f"matrix size {self.matrix.shape}")

        self.cut_intervals = cut_intervals
        self.interval_trees, self.chrBinBoundaries = \
            self.intervalListToIntervalTree(self.cut_intervals)

    def setMatrix(self, matrix, cut_intervals):
        """
        Initialize a matrix with a given matrix
        and cut_intervals. Mostly useful for
        testing.
        """

        # check that the matrix is squared
        if matrix.shape[0] != matrix.shape[1]:
            raise ValueError(f"Matrix is not squared. Shape is {matrix.shape}")
        if len(cut_intervals) != matrix.shape[0]:
            raise ValueError(f"Length of cut_intervals {len(cut_intervals)} does not match the matrix size {matrix.shape}")

        self.matrix = matrix
        self.cut_intervals = cut_intervals
        self.interval_trees, self.chrBinBoundaries = \
            self.intervalListToIntervalTree(self.cut_intervals)

    def getBinSize(self):
        """
        estimates the bin size. In case the bin size
        is not equal for all bins (maybe except for the
        bin at the en of the chromosomes) a warning is issued.
        In case of uneven bins, the median is returned.
        """
        if self.bin_size is None:
            chrom, start, end, extra = zip(*self.cut_intervals)
            diff = np.array(end) - np.array(start)
            # If there is only one bin:
            if len(diff) == 1:
                self.bin_size = diff[0]
                return self.bin_size
            # If there are more bins, the diff will be compared
            # to the median of the differences between starts
            median = int(np.median(np.concatenate([np.diff([start for chro, start, end, extra in self.cut_intervals if chro == cur_chrom]) for cur_chrom, nb in Counter(chrom).items() if nb > 1])))

            # check if the bin size is
            # homogeneous
            if len(np.flatnonzero(diff != median)) > (len(diff) * 0.01):
                self.bin_size_homogeneous = False
                if self.non_homogeneous_warning_already_printed is False:
                    log.warning('Bin size is not homogeneous. \
                                      Median %f\n', median)
                    self.non_homogeneous_warning_already_printed = True
            self.bin_size = median
        return self.bin_size

    def getMatrix(self):
        matrix = self.matrix.todense()
        if len(self.nan_bins):
            # to set NaN values the matrix type has to be
            # float. Corrected matrices are of float
            # type while uncorrected matrices are of
            # of int type
            if np.issubdtype(self.matrix, 'float') is False:
                matrix = matrix.astype(float)
            matrix[self.nan_bins, :] = np.nan
            matrix[:, self.nan_bins] = np.nan

        return matrix

    def getChrBinRange(self, chrName):
        """
        Given a chromosome name,
        This functions return the start and end bin indices in the matrix
        """

        if chrName in self.chrBinBoundaries:
            return self.chrBinBoundaries[chrName]
        raise ValueError(f"chrName: {chrName} not found in chrBinBoundaries"
                        f"valid chromosomes are: {self.chrBinBoundaries.keys()}")

    def getChrNames(self):
        """
        returns the names of the chromosomes
        present in the matrix
        """
        return list(self.chrBinBoundaries)

    def getBinPos(self, binIndex):
        """
        given a bin, it returns the chromosome name,
        start position and end position
        """
        if binIndex < len(self.cut_intervals):
            return self.cut_intervals[binIndex]
        raise ValueError(f"binIndex: {binIndex} not found")

    def getRegionBinRange(self, chrname, startpos, endpos):
        """
        Given a chromosome region, this function returns
        the bin indices that overlap with such region.
        """

        try:
            # chromosome_size = hic_matrix.get_chromosome_sizes()
            # chrname = check_chrom_str_bytes(self.interval_trees, chrname)
            if not isinstance(next(iter(self.interval_trees)), type(chrname)):
                if isinstance(next(iter(self.interval_trees)), str):
                    chrname = toString(chrname)
                elif isinstance(next(iter(self.interval_trees)), bytes):
                    chrname = toBytes(chrname)
                elif isinstance(next(iter(self.interval_trees)), np.bytes_):
                    chrname = toBytes(chrname)
            # chr_end_pos = chromosome_size[chrname]
            # self.interval_trees[chrname]
            if chrname not in self.interval_trees:
                raise ValueError(f"chromosome: {chrname} name not found in matrix"
                                 f"valid names are: {self.interval_trees.keys()}"
                                )
        except KeyError as ke:
            log.exception("chromosome: %s name not found in matrix", chrname)
            log.exception("valid names are: ")
            log.exception(self.interval_trees.keys())
            log.exception(str(ke))

        try:
            startpos = int(startpos)
            endpos = int(endpos)
        except ValueError as ve:
            log.exception("%d or %d  are not valid "
                          "position values.", startpos, endpos)
            log.exception(str(ve))

        try:

            startbin = sorted(self.interval_trees[chrname][startpos:startpos + 1])[0].data
            endbin = sorted(self.interval_trees[chrname][endpos:endpos + 1])[0].data
        except IndexError:
            # log.exception("chrname: " + chrname)
            # log.exception("len intervaltree: "+len(self.interval_trees[chrname]))
            # log.exception("start and end pos:" + startpos + ":::" + endpos )
            log.exception("Index error")
            return None

        return startbin, endbin

    @staticmethod
    def getDistList(rows, cols, cut_intervals):
        """
            Given a list of rows and cols
            an array is returned containing
            the genomic distance between
            each element of the row array
            with each element of the col array.
            -1 is returned for inter-chromosomal
            interactions.

            A matching list containing the chromosome name
            is also returned
        """
        chrnamelist, startlist, _, _ = zip(*cut_intervals)
        # now the distance between any two points
        # is computed and arranged such that for each
        # element of the data array, a corespondent distance is stored
        start_row = np.take(startlist, rows)
        start_col = np.take(startlist, cols)
        dist_list = start_col - start_row

        # now  all distances that are between chromosomes are removed
        # to do this I convert the array of chromosomes to
        # a array of indices. Then, when subtracting the
        # values that correspond to matrix.row and matrix.col
        # using the array of indices, any value other
        # than 0 means inter-chromosomal row,col combination.

        # chr_id_list is based on a trick using np.unique
        # to get from a list of strings
        # a list of integers
        chr_id_list = np.unique(chrnamelist, return_inverse=True)[1]

        chr_row = np.take(chr_id_list, rows)
        chr_col = np.take(chr_id_list, cols)
        chr_diff = chr_row - chr_col
        # set in dist_list array '-1' for all interchromosomal values
        dist_list[chr_diff != 0] = -1

        # make a corresponding chromosome name list
        # if filtering per chromosome is required
        chrom_list = np.take(chrnamelist, rows)
        chrom_list[chr_diff != 0] = ''

        return dist_list, chrom_list

    @staticmethod
    def fit_cut_intervals(cut_intervals):
        # check that the matrix has bins of same size
        # otherwise try to adjust the bins to
        # to match a regular binning
        if len(cut_intervals) <= 1:
            # do nothing if there is only one interval
            return cut_intervals
        chrom, start, end, extra = zip(*cut_intervals)

        median = int(np.median(np.concatenate([np.diff([start for chro, start, end, extra in cut_intervals if chro == cur_chrom]) for cur_chrom, nb in Counter(chrom).items() if nb > 1])))
        diff = np.array(end) - np.array(start)
        # check if the bin size is homogeneous
        if len(np.flatnonzero(diff != median)) > (len(diff) * 0.01):
            # set the start position of a bin to the closest multiple
            # of the median
            def snap_nearest_multiple(start_x, m):
                resi = [-1 * (start_x % m), -start_x % m]
                return start_x + resi[np.argmin(np.abs(resi))]
            start = [snap_nearest_multiple(x, median) for x in start]
            end = [snap_nearest_multiple(x, median) for x in end]
            cut_intervals = list(zip(chrom, start, end, extra))
            log.info('[getCountsByDistance] Bin size is not '
                     'homogeneous, setting \n'
                     'the bin distance to the median: %f\n', median)
        return cut_intervals

    def convert_to_zscore_matrix(self, maxdepth=None, perchr=False):
        return self.convert_to_obs_exp_matrix(maxdepth=maxdepth, zscore=True, perchr=perchr)

    def convert_to_obs_exp_matrix(self, maxdepth=None, zscore=False, perchr=False, pSkipTriu=False):
        """
        Converts a corrected counts matrix into a
        obs / expected matrix or z-scores fast.

        The caveat is that the obs/exp or z-score are only
        computed for non-zero values, although zero values that
        are not part of the sparse matrix are considered.

        For each diagonal the mean (and std when computing z-scores) are
        calculated and then each non-zero value of the sparse matrix is
        replaced by the obs/exp or z-score.

        Parameters
        ----------
        maxdepth: maximum distance from the diagonal to consider. All contacts beyond this distance will not
                         be considered.
        zscore: if a zscore wants to be returned instead of obs/exp


        Returns
        -------
        observed / expected sparse matrix


        nans occur where the standard deviation is zero
        """

        binsize = self.getBinSize()
        max_depth_in_bins = None

        if maxdepth:
            if maxdepth < binsize:
                raise ValueError(f"Please specify a maxDepth larger than bin size ({binsize})")

            max_depth_in_bins = int(float(maxdepth * 1.5) / binsize)
            # work only with the upper matrix
            # and remove all pixels that are beyond
            # max_depth_in_bis
            # (this is done by subtracting a second sparse matrix
            # that contains only the upper matrix that wants to be removed.
            if not pSkipTriu:
                self.matrix = triu(self.matrix, k=0, format='csr') - \
                    triu(self.matrix, k=max_depth_in_bins, format='csr')
        else:
            if not pSkipTriu:
                self.matrix = triu(self.matrix, k=0, format='csr')

        self.matrix.eliminate_zeros()
        depth = None
        if zscore is True:
            m_size = self.matrix.shape[0]
            if max_depth_in_bins is not None:
                depth = max_depth_in_bins
            else:
                depth = m_size
                estimated_size_dense_matrix = m_size ** 2 * 8
                if estimated_size_dense_matrix > 100e6:
                    log.info("To compute z-scores a dense matrix is required. This will use \n"
                             "%f Mb of memory.\n To reduce memory use the maxdeph option."
                             "", estimated_size_dense_matrix / 1e6)

            # to compute zscore the zero values need to be accounted and the matrix
            # need to become dense. This is only practical if only up to certain distance
            # wants to be evaluated, otherwise the dense matrix is too large.
            # To make the matrix dense and keep the same computations as when
            # the matrix is sparse the following is done:
            # A sparse diagonal matrix of shape = matrix.shape is created with ones
            # (only upper triangle contains diagonals up to maxdeph)
            # This  sparse matrix is then added to self.matrix
            # then, -1 is subtracted to the self.matrix.data, thus effectively
            # adding zeros.
            diag_mat_ones = diags(np.repeat([1], m_size * depth).reshape(depth, m_size), list(range(depth)))

            self.matrix += diag_mat_ones

        trasf_matrix = lil_matrix(self.matrix.shape)

        chr_submatrix = OrderedDict()
        cut_intervals = OrderedDict()
        chrom_sizes = OrderedDict()
        chrom_range = OrderedDict()
        if perchr:
            for chrname in self.getChrNames():
                chr_range = self.getChrBinRange(chrname)
                chr_submatrix[chrname] = self.matrix[chr_range[0]:chr_range[1], chr_range[0]:chr_range[1]].tocoo()
                cut_intervals[chrname] = [self.cut_intervals[x] for x in range(chr_range[0], chr_range[1])]
                chrom_sizes[chrname] = [chr_submatrix[chrname].shape[0]]
                chrom_range[chrname] = (chr_range[0], chr_range[1])

        else:
            chr_submatrix['all'] = self.matrix.tocoo()
            cut_intervals['all'] = self.cut_intervals
            # chrom_sizes['all'] = np.array([v[1] - v[0] for k, v in iteritems(self.chrBinBoundaries)])
            chrom_sizes['all'] = np.array([v[1] - v[0] for k, v in self.chrBinBoundaries.items()])

            chrom_range['all'] = (0, self.matrix.shape[0])

        # for chrname, submatrix in iteritems(chr_submatrix):
        for chrname, submatrix in chr_submatrix.items():

            log.info("processing chromosome %s\n", chrname)
            if zscore is True:
                # this step has to be done after tocoo()
                submatrix.data -= 1

            dist_list, _ = self.getDistList(submatrix.row, submatrix.col,
                                                     hiCMatrix.fit_cut_intervals(cut_intervals[chrname]))

            # to get the sum of all values at a given distance I use np.bincount which
            # is quite fast. However, the input of bincount is positive integers. Moreover
            # it returns the sum for every consecutive integer, even if this is not on the list.
            # Thus, dist_list, which contains the distance in bp between any two bins is
            # converted to bin distance.

            # Because positive integers are needed we add +1 to all bin distances
            # such that the value of -1 (which means different chromosomes) can now be used

            dist_list[dist_list == -1] = -binsize  # pylint: disable=E1130
            # divide by binsize to get a list of bin distances and add +1 to remove negative values
            dist_list = (np.array(dist_list).astype(float) / binsize).astype(int) + 1

            # for each distance, return the sum of all values
            sum_counts = np.bincount(dist_list, weights=submatrix.data)
            distance_len = np.bincount(dist_list)
            # compute the average for each distance
            mat_size = submatrix.shape[0]
            mu = {}
            std = {}
            # compute mean value for each distance

            for bin_dist_plus_one, sum_value in enumerate(sum_counts):
                if maxdepth and bin_dist_plus_one == 0:  # this is for intra chromosomal counts
                    # when max depth is set, the computation
                    # of the total_intra is not accurate and is safer to
                    # output np.nan
                    mu[bin_dist_plus_one] = np.nan
                    std[bin_dist_plus_one] = np.nan
                    continue

                if bin_dist_plus_one == 0:
                    total_intra = mat_size ** 2 - sum([size ** 2 for size in chrom_sizes[chrname]])
                    diagonal_length = int(total_intra / 2)
                else:
                    # to compute the average counts per distance we take the sum_counts and divide
                    # by the number of values on the respective diagonal
                    # which is equal to the size of each chromosome - the diagonal offset (for those
                    # chromosome larger than the offset)
                    # In the following example with two chromosomes
                    # the first (main) diagonal has a size equal to the matrix (6),
                    # while the next has 1 value less for each chromosome (4) and the last one has only 2 values

                    # 0 1 2 . . .
                    # - 0 1 . . .
                    # - - 0 . . .
                    # . . . 0 1 2
                    # . . . - 0 1
                    # . . . - - 0

                    # idx - 1 because earlier the values where
                    # shifted.
                    diagonal_length = sum([size - (bin_dist_plus_one - 1) for size in chrom_sizes[chrname] if size > (bin_dist_plus_one - 1)])
                    log.debug("Type of diagonal_length %s", type(diagonal_length))

                # the diagonal length should contain the number of values at a certain distance.
                # If the matrix is dense, the distance_len[bin_dist_plus_one] correctly contains the number of values
                # If the matrix is equally spaced, then, the diagonal_length as computed before is accurate.
                # But, if the matrix is both sparse and with unequal bins, then none of the above methods is
                # accurate but the the diagonal_length as computed before will be closer.
                diagonal_length = max(diagonal_length, distance_len[bin_dist_plus_one])
                log.debug("Type of diagonal_length %s", type(diagonal_length))

                if diagonal_length == 0:
                    mu[bin_dist_plus_one] = np.nan
                else:
                    mu[bin_dist_plus_one] = np.float64(sum_value) / diagonal_length

                if np.isnan(sum_value):
                    log.info("nan value found for distance %f\n", (bin_dist_plus_one - 1) * binsize)

                # if zscore is needed, compute standard deviation: std = sqrt(mean(abs(x - x.mean())**2))
                if zscore:
                    values_sqrt_diff = \
                        np.abs((submatrix.data[dist_list == bin_dist_plus_one] - mu[bin_dist_plus_one]) ** 2)
                    # the standard deviation is the sum of the differences with mu squared (value variable)
                    # plus all zeros that are not included in the sparse matrix
                    # for which the standard deviation is
                    # (0 - mu)**2 = (mu)**2
                    # The number of zeros is the diagonal length - the length of the non zero values
                    zero_values_sqrt_diff_sum = (diagonal_length - len(values_sqrt_diff)) * mu[bin_dist_plus_one] ** 2

                    _std = np.sqrt((values_sqrt_diff.sum() + zero_values_sqrt_diff_sum) / diagonal_length)
                    std[bin_dist_plus_one] = _std

            # use the expected values to compute obs/exp
            transf_ma = np.zeros(len(submatrix.data))
            for idx, value in enumerate(submatrix.data):
                if depth is not None and dist_list[idx] > depth + 1:
                    continue
                if zscore:
                    if std[dist_list[idx]] == 0:
                        transf_ma[idx] = np.nan
                    else:
                        transf_ma[idx] = (value - mu[dist_list[idx]]) / std[dist_list[idx]]
                else:
                    transf_ma[idx] = value / mu[dist_list[idx]]

            submatrix.data = transf_ma
            trasf_matrix[chrom_range[chrname][0]:chrom_range[chrname][1], chrom_range[chrname][0]:chrom_range[chrname][1]] = submatrix.tolil()

        self.matrix = trasf_matrix.tocsr()

        return self.matrix

    @staticmethod
    def dist_list_to_dict(data, dist_list):
        """
        splits data, into numeric groups defined by dist_list
        Return a dictionary containing, for
        each unique distance a dictionary
        """

        order = np.argsort(dist_list)
        dist_list = dist_list[order]
        data = data[order]

        # having the dist_list sorted, np.split
        # is used to divide the data into
        # groups that lie at the same distance, for this
        # np.diff together with np.flatnonzero is used to
        # find the indices where the distance changes.
        # the '+1' is needed because the np.diff array is
        # one element smaller than the original array, thus
        # the indices based no the np.diff array are off by 1
        # with respect to the original array
        groups = np.split(data, np.flatnonzero(np.diff(dist_list)) + 1)

        # because the dist_list is sorted
        # the order of the unique values
        # corresponds to that of the groups.
        # In other words, group[0]
        # has distance_unique[0]
        # np.sort after np.unique  in theory
        # is not needed, but just in case...
        distance_unique = np.sort(np.unique(dist_list))

        # convert to dictionary having as key
        # the distance
        distance = {}
        for index, d in enumerate(distance_unique):
            distance[d] = groups[index]

        return distance

    def keepOnlyTheseChr(self, chromosome_list):
        """
        given a list of chromosome names,
        these are kept, while any other is removed
        from the matrix
        """
        chromosome_list = check_chrom_str_bytes(self.interval_trees, chromosome_list)

        try:
            [self.chrBinBoundaries[x] for x in chromosome_list]
        except KeyError as e:
            raise ValueError(f"Chromosome name not in matrix. {str(e)}") from e

        self.restoreMaskedBins()
        size = self.matrix.shape
        # initialize a 1D array containing the columns (and rows) to
        # select. By default none are selected
        sel = np.empty(size[0], dtype=bool)
        sel[:] = False

        for chrName in list(self.interval_trees):
            if chrName not in chromosome_list:
                continue

            # identify start and end rows
            # of chromosomes that wants to be
            # kept
            index_start, index_end = self.getChrBinRange(chrName)
            sel[index_start:index_end] = True

        sel_id = np.flatnonzero(sel)
        mat = self.matrix[sel_id, :][:, sel_id]

        # update bin ids
        self.cut_intervals = [self.cut_intervals[x] for x in sel_id]

        # update correction factors
        if self.correction_factors is not None:
            self.correction_factors = [self.correction_factors[x] for x in sel_id]

        # keep track of nan bins
        if len(self.nan_bins):
            _temp = np.zeros(size[0])
            _temp[self.nan_bins] = 1
            _temp = _temp[sel_id]
            self.nan_bins = np.flatnonzero(_temp == 1)
        else:
            self.nan_bins = []

        self.numCols = len(sel_id)  # pylint: disable=W0201

        self.interval_trees, self.chrBinBoundaries = \
            self.intervalListToIntervalTree(self.cut_intervals)
        # remove distanceCounts
        try:
            self.distance_counts = None
        except AttributeError:
            pass
        self.matrix = mat
        return self.matrix

    def diagflat(self, value=np.nan):
        """
        sets
        the matrix diagonal to np.nan
        """
        M = self.matrix.shape[0]
        diagmatrix = dia_matrix((np.repeat(value, M), 0), shape=(M, M))
        self_diag = dia_matrix(([self.matrix.diagonal()], [0]), shape=(M, M))
        # take matrix, subtract the values of the diagonal such that
        # it becomes all zeros, replace with new values by adding them
        self.matrix = self.matrix - self_diag + diagmatrix
        return self.matrix

    def filterOutInterChrCounts(self):
        """
        set all inter chromosomal counts to np.nan
        """

        ma_coo = self.matrix.tocoo()
        dist_list, _ = hiCMatrix.getDistList(ma_coo.row, ma_coo.col,
                                             self.cut_intervals)

        # set to zero all cases in which dist_list is zero
        ma_coo.data[dist_list == -1] = 0

        self.matrix = ma_coo.tocsr()
        self.matrix.eliminate_zeros()
        return self.matrix

    def setMatrixValues(self, newMatrix):
        """
        replace the current matrix values
        by the given matrix values. The
        shapes have to coincide
        """
        assert self.matrix.shape == newMatrix.shape, \
            "Given matrix has different shape. New " \
            "values need to have the same shape as previous matrix."

        self.matrix = csr_matrix(newMatrix)

    def setCorrectionFactors(self, correction_factors):
        assert len(correction_factors) == self.matrix.shape[0], \
            "length of correction factors and length of matrix are different."
        self.correction_factors = correction_factors

    def reorderChromosomes(self, new_chr_order):
        new_order = []
        new_chr_order = check_chrom_str_bytes(self.chrBinBoundaries, new_chr_order)

        for chrName in new_chr_order:
            # check that the chromosome names are valid
            if chrName not in self.chrBinBoundaries:
                raise ValueError(f"Chromosome name '{chrName}' not found. Please check the correct spelling "
                                 "of the chromosomes and try again")
            orig = self.chrBinBoundaries[chrName]
            new_order.extend(list(range(orig[0], orig[1])))
        self.reorderBins(new_order)

    def reorderBins(self, new_order):
        """
        reorders the rows and colums of the
        matrix according to the new order.
        The new order can be smaller
        than the original matrix. In that
        case, the ids not in the
        new order are removed.
        """
        orig_num_rows = self.matrix.shape[0]
        self.matrix = self.matrix[new_order, :][:, new_order]
        self.cut_intervals = [self.cut_intervals[x] for x in new_order]
        # reorder the masked bins
        # keep track of nan bins
        if len(self.nan_bins):
            _temp = np.zeros(orig_num_rows)
            _temp[self.nan_bins] = 1
            _temp = _temp[new_order]
            self.nan_bins = np.flatnonzero(_temp == 1)
        else:
            self.nan_bins = []

        self.interval_trees, self.chrBinBoundaries = \
            self.intervalListToIntervalTree(self.cut_intervals)

    def maskChromosomes(self, pChromosomeList):
        mask_ids = []
        pChromosomeList = check_chrom_str_bytes(self.chrBinBoundaries, pChromosomeList)

        for chromosome in pChromosomeList:
            # check that the chromosome names are valid
            if chromosome not in self.chrBinBoundaries:
                raise ValueError(f"Chromosome name '{chromosome}' not found. Please check the correct spelling "
                                 "of the chromosomes and try again")
            orig = self.chrBinBoundaries[chromosome]
            mask_ids.extend(list(range(orig[0], orig[1])))
        self.maskBins(mask_ids)

    def maskBins(self, bin_ids=None):
        """
        Mask the list of bins given. Mask means
        to remove the bins from the matrix,
        and keep the information about the intervals
        as masked
        """
        # print("self.cut_intervalsMASKBINS___START", self.cut_intervals)

        if bin_ids is None or len(bin_ids) == 0:
            return
        self.printchrtoremove(bin_ids, restore_masked_bins=False)
        try:
            # check if a masked bin already exists
            if len(self.orig_bin_ids) > 0:
                M = self.matrix.shape[0]
                previous_bin_ids = self.orig_bin_ids[M:]
                # merge new and old masked bins
                bin_ids = np.unique(np.concatenate([previous_bin_ids, self.orig_bin_ids[bin_ids]]))
                np.sort(bin_ids)
                self.restoreMaskedBins()
        except Exception:  # pylint: disable=W0718
            pass

        # join with existing nan_bins
        if self.nan_bins is not None and len(self.nan_bins) > 0:
            log.info("found existing %d nan bins that will be "
                     "included for masking ", len(self.nan_bins))
            bin_ids = np.unique(np.concatenate([self.nan_bins, bin_ids]))
            self.nan_bins = []
        rows = cols = np.delete(list(range(self.matrix.shape[1])), bin_ids)

        self.matrix = self.matrix[rows, :][:, cols]

        # to keep track of removed bins
        # I add their ids to the end of the rows vector
        # to reverse the changes, I just need to do an argsort
        # to put the removed bins in place
        # log.debug("bins_ids {}".format(bin_ids))
        self.orig_bin_ids = np.concatenate([rows, bin_ids])

        new_cut_intervals = [self.cut_intervals[x] for x in rows]

        self.orig_cut_intervals = new_cut_intervals + [self.cut_intervals[x] for x in bin_ids]

        self.cut_intervals = new_cut_intervals

        self.interval_trees, self.chrBinBoundaries = self.intervalListToIntervalTree(self.cut_intervals)

        if self.correction_factors is not None:
            self.correction_factors = self.correction_factors[rows]

    def update_matrix(self, new_matrix, new_cut_intervals):
        """
        give a new matrix and list of cut intervals, the matrix,  cut intervals and
        the respective tree are updated
        :param new_matrix: now values for the sparse matrix
        :param new_cut_intervals: list of cut intervals, each entry being a tuple of the form
        (chrom, start, end, coverage)
        :return:
        """
        if len(self.orig_bin_ids) > 0:
            raise ValueError("matrix contains masked bins. Restore masked bins first")

        assert len(new_cut_intervals) == new_matrix.shape[0], "matrix shape and len of cut intervals do not match"

        self.matrix = new_matrix
        self.cut_intervals = new_cut_intervals

        self.interval_trees, self.chrBinBoundaries = \
            self.intervalListToIntervalTree(self.cut_intervals)

        self.nan_bins = np.flatnonzero(self.matrix.sum(0).A == 0)

    def restoreMaskedBins(self):
        """
        Puts backs into the matrix the bins
        removed
        """
        if len(self.orig_bin_ids) == 0:
            return
        # the rows to add are
        # as an empty sparse matrix
        M = self.matrix.shape[0]
        N = len(self.orig_bin_ids) - M
        rows_mat = csr_matrix((N, M))
        # cols to add
        cols_mat = csr_matrix((M + N, N))

        # add the rows and cols at the end of the
        # current matrix
        self.matrix = sparse_vstack([self.matrix, rows_mat])
        self.matrix = sparse_hstack([self.matrix, cols_mat], format='csr')

        # the new matrix has the right number of cols and rows, now
        # they need to be reordered to be back in their original places
        rows = cols = np.argsort(self.orig_bin_ids)
        self.matrix = self.matrix[rows, :][:, cols]
        self.cut_intervals = [self.orig_cut_intervals[x] for x in rows]
        self.interval_trees, self.chrBinBoundaries = \
            self.intervalListToIntervalTree(self.cut_intervals)
        # set as nan_bins the masked bins that were restored
        self.nan_bins = self.orig_bin_ids[M:]

        if self.correction_factors is not None:
            # add missing values as nans at end of array
            self.correction_factors = np.concatenate([self.correction_factors,
                                                      np.repeat(np.nan, N)])
            # reorder array
            self.correction_factors = self.correction_factors[rows]

        # reset orig bins ids and cut intervals
        self.orig_bin_ids = []
        self.orig_cut_intervals = []
        log.info("masked bins were restored\n")

    def reorderMatrix(self, orig, dest):
        """
        Given a matrix, a region over the diagonal is moved from
        its origin to a new destination. With this method a
        new order of the chromosomes can be produced.
        :param orig: a tuple containing the indices of the region to be moved
        :param dest: the index of the region into which to insert
                     the section moved
        """

        rows = np.delete(list(range(self.matrix.shape[1])), range(orig[0], orig[1]))

        if dest > orig[1]:
            dest = dest - (orig[1] - orig[0])

        rows = cols = np.insert(
            rows, np.repeat(dest, orig[1] - orig[0]), list(range(orig[0], orig[1])))
        self.matrix = self.matrix[rows, :][:, cols]
        self.cut_intervals = [self.cut_intervals[x] for x in rows]
        self.interval_trees, self.chrBinBoundaries = \
            self.intervalListToIntervalTree(self.cut_intervals)

        if self.correction_factors is not None:
            self.correction_factors = self.correction_factors[rows]

    def truncTrans(self, high=0.05):
        """Truncates trans contacts to remove blowouts
        Clip high counts in trans regions (i.e. between
        chromosomes) to the max value found in the 1-high*100
        percentile

        :param:  high : float, 0<high<1, optional
            Fraction of top trans interactions to be removed
        """
        mat = self.matrix.tocoo()
        dist_list = hiCMatrix.getDistList(mat.row, mat.col, self.cut_intervals)
        if np.count_nonzero(dist_list == -1) == 0:
            return
        max_inter = np.percentile(mat.data[dist_list == -1], (100 - high))
        # This needs to be fixed I guess Joachim wanted to put '=' instead of '=='
        mat.data[(mat.data >= max_inter) & (dist_list == -1)] == max_inter  # pylint: disable=W0104

        self.setMatrixValues(mat)

    def printchrtoremove(self, to_remove, label="Number of poor regions to remove", restore_masked_bins=True):
        """
        prints out the number of bin per chromosomes
        that will be removed
        """
        cnt = {}
        try:
            self.prev_to_remove
        except Exception:  # pylint: disable=W0718
            log.debug("No self.prev_to_remove defined, defining it now.")
            self.prev_to_remove = np.array([])  # pylint: disable=W0201

        # if the same information was already printed don't
        # show it again.
        if np.array_equal(self.prev_to_remove, to_remove):
            return

        if restore_masked_bins:
            try:
                # check if a masked bin already exists
                if len(self.orig_bin_ids) > 0:
                    log.info("Masked bins already present")
                    self.restoreMaskedBins()
            except Exception:  # pylint: disable=W0718
                pass
        for idx in to_remove:
            chrom = self.cut_intervals[idx][0]
            if chrom not in cnt:
                cnt[chrom] = 0
            cnt[chrom] += 1

        log.info('%s: %d %s', label, len(to_remove), cnt)
        self.prev_to_remove = to_remove  # pylint: disable=W0201

    def get_chromosome_sizes_real(self):
        '''
        Function returns the size of a chromosome as it is stored in the  matrix.
        The size can differ if e.g. some area from the start or end of a chromosome is not present in the interaction matrix.
        '''
        if self.chrBinBoundaries and len(self.chrBinBoundaries) > 0:
            chrom_sizes = OrderedDict()
            # for chrom, (start_bin, end_bin) in iteritems(self.chrBinBoundaries):
            for chrom, (start_bin, end_bin) in self.chrBinBoundaries.items():
                chrom, start0, _, _ = self.cut_intervals[start_bin]
                chrom, _, end1, _ = self.cut_intervals[end_bin - 1]
                chrom_sizes[chrom] = end1 - start0 + 1

            return chrom_sizes
        return None

    def get_chromosome_sizes(self):
        '''
        Function returns the size of a chromosome as it is stored in the  matrix, assuming the chromosome starts is always at its genomic position 0.
        '''
        if self.chrBinBoundaries and len(self.chrBinBoundaries) > 0:
            chrom_sizes = OrderedDict()
            # for chrom, (start_bin, end_bin) in iteritems(self.chrBinBoundaries):
            for chrom, (_, end_bin) in self.chrBinBoundaries.items():

                chrom, _, end, _ = self.cut_intervals[end_bin - 1]
                chrom_sizes[chrom] = end

            return chrom_sizes
        return None

    def intervalListToIntervalTree(self, interval_list):
        """
        given an ordered list of (chromosome name, start, end)
        this is transformed to a number of interval trees,
        one for each chromosome
        """
        cut_int_tree = {}
        chrbin_boundaries = OrderedDict()
        if len(interval_list) == 0:
            log.warning("Interval list is empty")
            return cut_int_tree, chrbin_boundaries

        intval_id = 0
        chr_start_id = 0
        previous_chrom = None
        for intval in interval_list:
            chrom, start, end = intval[0:3]
            start = int(start)
            end = int(end)
            if previous_chrom != chrom:
                if previous_chrom is None:
                    previous_chrom = chrom

                chrbin_boundaries[previous_chrom] = \
                    (chr_start_id, intval_id)
                chr_start_id = intval_id
                cut_int_tree[chrom] = IntervalTree()
                previous_chrom = chrom

            cut_int_tree[chrom].add(Interval(start, end, intval_id))

            intval_id += 1
        chrbin_boundaries[chrom] = (chr_start_id, intval_id)

        return cut_int_tree, chrbin_boundaries


def check_cooler(pFileName):
    if pFileName.endswith('.cool') or '.mcool' in pFileName:
        if cooler.fileops.is_cooler(pFileName):
            return True
    return False




import logging

from scipy.sparse import csr_matrix

from .matrixFile import MatrixFile

log = logging.getLogger(__name__)


class Hicpro(MatrixFile):

    def __init__(self, pMatrixFile, pBedFile):
        super().__init__(pMatrixFileName=pMatrixFile, pBedFile=pBedFile)

    def load(self):
        instances = []
        features = []
        data = []
        with open(self.matrixFileName, 'r', encoding="utf-8") as matrix_file:
            for line in matrix_file:
                x, y, value = line.strip().split('\t')
                instances.append(int(x) - 1)
                features.append(int(y) - 1)
                data.append(float(value))
        cut_intervals = []
        with open(self.bedFile, 'r', encoding="utf-8") as bed_file:
            for line in bed_file:
                chrom, start, end, value = line.strip().split('\t')
                cut_intervals.append((chrom, int(start), int(end), int(value)))

        shape = len(cut_intervals)

        matrix = csr_matrix((data, (instances, features)), shape=(shape, shape))

        nan_bins = None
        distance_counts = None
        correction_factors = None
        return matrix, cut_intervals, nan_bins, distance_counts, correction_factors

    def save(self, pFileName, pSymmetric=None, pApplyCorrection=None):
        self.matrix.eliminate_zeros()
        instances, features = self.matrix.nonzero()
        data = self.matrix.data

        with open(pFileName, 'w', encoding="utf-8") as matrix_file:
            for x, y, value in zip(instances, features, data):
                matrix_file.write(str(int(x + 1)) + '\t' + str(int(y + 1)) + '\t' + str(value) + '\n')

        with open(self.bedFile, 'w', encoding="utf-8") as bed_file:
            for i, interval in enumerate(self.cut_intervals):
                bed_file.write('\t'.join(map(str, interval[:3])) + '\t' + str(i + 1) + '\n')



import logging

import gzip

from scipy.sparse import csr_matrix

from hicmatrix.utilities import opener

from .matrixFile import MatrixFile

log = logging.getLogger(__name__)


class Homer(MatrixFile):

    def __init__(self, pMatrixFile):
        super().__init__(pMatrixFile)

    def load(self):
        cut_intervals = []

        # matrix_file = opener(self.matrixFileName)
        with opener(self.matrixFileName) as matrix_file:
            values = matrix_file.readline()
            values = values.strip().split(b'\t')

            # get bin size
            start_first = int(values[2].strip().split(b'-')[1])
            start_second = int(values[3].strip().split(b'-')[1])
            bin_size = start_second - start_first
            for value in values[2:]:
                chrom, start = value.strip().split(b'-')
                cut_intervals.append((chrom.decode('ascii'), int(start), int(start) + bin_size, 1))

            matrix_dense = []
            for line in matrix_file:
                values = line.split(b'\t')
                data = []
                for value in values[2:]:
                    data.append(float(value))
                matrix_dense.append(data)
        # matrix_file.close()
        matrix = csr_matrix(matrix_dense)
        nan_bins = None
        distance_counts = None
        correction_factors = None
        return matrix, cut_intervals, nan_bins, distance_counts, correction_factors

    def save(self, pFileName, pSymmetric=None, pApplyCorrection=None):

        with gzip.open(pFileName, 'wt') as homerMatrixFile:
            homerMatrixFile.write('HiCMatrix (directory=.)\tRegions\t')
            for bin_interval in self.cut_intervals:
                homerMatrixFile.write(f'{bin_interval[0]}-{bin_interval[1]}\t')
            homerMatrixFile.write('\n')

            for i in range(self.matrix.shape[0]):
                data = '\t'.join(map(str, self.matrix[i, :].toarray().flatten()))
                homerMatrixFile.write(f'{self.cut_intervals[i][0]}-{self.cut_intervals[i][1]}\t{self.cut_intervals[i][0]}-{self.cut_intervals[i][1]}\t')
                homerMatrixFile.write(f'{data}')
                if i < self.matrix.shape[0] - 1:
                    homerMatrixFile.write('\n')



import logging
import os
from os import unlink

import numpy as np
import tables
from scipy.sparse import csr_matrix, triu

from hicmatrix.utilities import toString

from .matrixFile import MatrixFile

log = logging.getLogger(__name__)


class H5(MatrixFile):

    def __init__(self, pMatrixFile):
        super().__init__(pMatrixFile)

    def load(self):
        """
        Loads a matrix stored in h5 format
        :param matrix_filename:
        :return: matrix, cut_intervals, nan_bins, distance_counts, correction_factors
        """
        log.debug('Load in h5 format')

        with tables.open_file(self.matrixFileName, 'r') as f:
            parts = {}
            try:
                for matrix_part in ('data', 'indices', 'indptr', 'shape'):
                    parts[matrix_part] = getattr(f.root.matrix, matrix_part).read()
            except Exception as e:  # pylint: disable=W0718
                log.info('No h5 file. Please check parameters concerning the file type!')
                # Should probably be raise e:
                e  # pylint: disable=W0104
            matrix = csr_matrix(tuple([parts['data'], parts['indices'], parts['indptr']]),
                                shape=parts['shape'])
            # matrix = hiCMatrix.fillLowerTriangle(matrix)
            # get intervals
            intvals = {}
            for interval_part in ('chr_list', 'start_list', 'end_list', 'extra_list'):
                if toString(interval_part) == toString('chr_list'):
                    chrom_list = getattr(f.root.intervals, interval_part).read()
                    intvals[interval_part] = toString(chrom_list)
                else:
                    intvals[interval_part] = getattr(f.root.intervals, interval_part).read()

            cut_intervals = list(zip(intvals['chr_list'], intvals['start_list'], intvals['end_list'], intvals['extra_list']))
            assert len(cut_intervals) == matrix.shape[0], \
                f"Error loading matrix. Length of bin intervals ({len(cut_intervals)}) is different than the " \
                f"size of the matrix ({matrix.shape[0]})"

            # get nan_bins
            try:
                if hasattr(f.root, 'nan_bins'):
                    nan_bins = f.root.nan_bins.read()
                else:
                    nan_bins = np.array([])
            except Exception:  # pylint: disable=W0718
                nan_bins = np.array([])

            # get correction factors
            try:
                if hasattr(f.root, 'correction_factors'):
                    correction_factors = f.root.correction_factors.read()
                    assert len(correction_factors) == matrix.shape[0], \
                        "Error loading matrix. Length of correction factors does not" \
                        "match size of matrix"
                    correction_factors = np.array(correction_factors)
                    mask = np.isnan(correction_factors)
                    correction_factors[mask] = 0
                    mask = np.isinf(correction_factors)
                    correction_factors[mask] = 0
                else:
                    correction_factors = None
            except Exception:  # pylint: disable=W0718
                correction_factors = None

            try:
                # get correction factors
                if hasattr(f.root, 'distance_counts'):
                    distance_counts = f.root.correction_factors.read()
                else:
                    distance_counts = None
            except Exception:  # pylint: disable=W0718
                distance_counts = None
            return matrix, cut_intervals, nan_bins, distance_counts, correction_factors

    def save(self, pFileName, pSymmetric=True, pApplyCorrection=None):
        """
        Saves a matrix using hdf5 format
        :param pFileName:
        :return: None
        """
        log.debug('Save in h5 format')

        # self.restoreMaskedBins()
        if not pFileName.endswith(".h5"):
            pFileName += ".h5"

        # if the file name already exists
        # try to find a new suitable name
        if os.path.isfile(pFileName):
            log.warning("*WARNING* File already exists %s\n "
                        "Overwriting ...\n", pFileName)

            unlink(pFileName)
        if self.nan_bins is None:
            self.nan_bins = np.array([])
        elif not isinstance(self.nan_bins, np.ndarray):
            self.nan_bins = np.array(self.nan_bins)

        # save only the upper triangle of the
        if pSymmetric:
            # symmetric matrix
            matrix = triu(self.matrix, k=0, format='csr')
        else:
            matrix = self.matrix
        matrix.eliminate_zeros()

        filters = tables.Filters(complevel=5, complib='blosc')
        with tables.open_file(pFileName, mode="w", title="HiCExplorer matrix") as h5file:
            matrix_group = h5file.create_group("/", "matrix", )
            # save the parts of the csr matrix
            for matrix_part in ('data', 'indices', 'indptr', 'shape'):
                arr = np.array(getattr(matrix, matrix_part))
                atom = tables.Atom.from_dtype(arr.dtype)
                ds = h5file.create_carray(matrix_group, matrix_part, atom,
                                          shape=arr.shape,
                                          filters=filters)
                ds[:] = arr

            # save the matrix intervals
            intervals_group = h5file.create_group("/", "intervals", )
            chr_list, start_list, end_list, extra_list = zip(*self.cut_intervals)  # pylint: disable=W0612
            for interval_part in ('chr_list', 'start_list', 'end_list', 'extra_list'):
                arr = np.array(eval(interval_part))  # pylint: disable=W0123
                atom = tables.Atom.from_dtype(arr.dtype)
                ds = h5file.create_carray(intervals_group, interval_part, atom,
                                          shape=arr.shape,
                                          filters=filters)
                ds[:] = arr

            # save nan bins
            if len(self.nan_bins):
                atom = tables.Atom.from_dtype(self.nan_bins.dtype)
                ds = h5file.create_carray(h5file.root, 'nan_bins', atom,
                                          shape=self.nan_bins.shape,
                                          filters=filters)
                ds[:] = self.nan_bins

            # save corrections factors
            if self.correction_factors is not None and len(self.correction_factors):
                self.correction_factors = np.array(self.correction_factors)
                mask = np.isnan(self.correction_factors)
                self.correction_factors[mask] = 0
                atom = tables.Atom.from_dtype(self.correction_factors.dtype)
                ds = h5file.create_carray(h5file.root, 'correction_factors', atom,
                                          shape=self.correction_factors.shape,
                                          filters=filters)
                ds[:] = np.array(self.correction_factors)

            # save distance counts
            if self.distance_counts is not None and len(self.distance_counts):
                atom = tables.Atom.from_dtype(self.distance_counts.dtype)
                ds = h5file.create_carray(h5file.root, 'distance_counts', atom,
                                          shape=self.distance_counts.shape,
                                          filters=filters)
                ds[:] = np.array(self.distance_counts)



import gzip
import sys

import numpy as np


def toString(s):
    """
    This takes care of python2/3 differences
    """
    if isinstance(s, str):
        return s
    if isinstance(s, bytes):  # or isinstance(s, np.bytes_):
        if sys.version_info[0] == 2:
            return str(s)
        return s.decode('ascii')
    if isinstance(s, list):
        return [toString(x) for x in s]
    if isinstance(s, np.ndarray):
        return s.astype(str)
    return s


def toBytes(s):
    """
    Like toString, but for functions requiring bytes in python3
    """
    if sys.version_info[0] == 2:
        return s
    if isinstance(s, bytes):
        return s
    # if isinstance(s, np.bytes_):
    #     return np.bytes_(s)
    if isinstance(s, str):
        return bytes(s, 'ascii')
    if isinstance(s, list):
        return [toBytes(x) for x in s]
    return s


def check_chrom_str_bytes(pIteratableObj, pObj):
    # determine type
    if isinstance(pObj, list) and len(pObj) > 0:
        type_ = type(pObj[0])
    else:
        type_ = type(pObj)
    if not isinstance(type(next(iter(pIteratableObj))), type_):
        if isinstance(next(iter(pIteratableObj)), str):
            pObj = toString(pObj)
        elif type(next(iter(pIteratableObj))) in [bytes, np.bytes_]:
            pObj = toBytes(pObj)
    return pObj


def convertNansToZeros(ma):
    nan_elements = np.flatnonzero(np.isnan(ma.data))
    if len(nan_elements) > 0:
        ma.data[nan_elements] = 0.0
    return ma


def convertNansToOnes(pArray):
    nan_elements = np.flatnonzero(np.isnan(pArray))
    if len(nan_elements) > 0:
        pArray[nan_elements] = 1.0
    return pArray


def enlarge_bins(bin_intervals):
    r"""
    takes a list of consecutive, but not
    directly touching, bin intervals
    and joins them such that the
    end and start of consecutive bins
    is the same.

    >>> bin_intervals = [('chr1', 10, 50, 1), ('chr1', 50, 80, 2),
    ... ('chr2', 10, 60, 3), ('chr2', 70, 90, 4)]
    >>> enlarge_bins(bin_intervals)
    [('chr1', 0, 50, 1), ('chr1', 50, 80, 2), ('chr2', 0, 65, 3), ('chr2', 65, 90, 4)]
    """
    # enlarge remaining bins
    chr_start = True
    for idx in range(len(bin_intervals) - 1):
        chrom, start, end, extra = bin_intervals[idx]
        chrom_next, start_next, end_next, extra_next = bin_intervals[idx + 1]

        if chr_start is True:
            start = 0
            chr_start = False
            bin_intervals[idx] = (chrom, start, end, extra)
        if chrom == chrom_next and end != start_next:
            middle = start_next - int((start_next - end) / 2)
            bin_intervals[idx] = (chrom, start, middle, extra)
            bin_intervals[idx + 1] = (chrom, middle, end_next, extra_next)
        if chrom != chrom_next:
            chr_start = True

    chrom, start, end, extra = bin_intervals[-1]
    bin_intervals[-1] = (chrom, start, end, extra)

    return bin_intervals


def opener(filename):
    """
    Determines if a file is compressed or not
    """
    f = open(filename, 'rb')  # pylint: disable=R1732
    # print("gzip or not?", f.read(2))

    if f.read(2) == b'\x1f\x8b':
        f.seek(0)
        return gzip.GzipFile(fileobj=f)

    f.seek(0)
    return f



import logging
import os

import gc
try:
    from importlib.metadata import version
except ImportError:
    from importlib_metadata import version

import cooler
import h5py
import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix, dok_matrix, lil_matrix, triu

from hicmatrix.utilities import (convertNansToOnes, toString)

from .matrixFile import MatrixFile

log = logging.getLogger(__name__)

class Cool(MatrixFile):

    def __init__(self, pMatrixFile=None):
        super().__init__(pMatrixFile)
        self.chrnameList = None
        self.correctionFactorTable = 'weight'
        self.correctionOperator = None
        self.enforceInteger = False
        self.appendData = False
        self.fileWasH5 = False
        self.applyCorrectionLoad = True
        self.hic_metadata = {}
        self.cool_info = None

        self.hic2cool_version = None
        self.hicmatrix_version = None
        self.distance = None
        self.matrixFormat = None
        self.matrixOnly = False
        self.noCutIntervals = False

    def getInformationCoolerBinNames(self):
        return cooler.Cooler(self.matrixFileName).bins().columns.values

    def load(self):
        log.debug('Load in cool format')
        self.minValue = None  # pylint: disable=W0201
        self.maxValue = None  # pylint: disable=W0201
        if self.matrixFileName is None:
            log.warning('No matrix is initialized')
        try:
            cooler_file = cooler.Cooler(self.matrixFileName)
            # if 'metadata' in cooler_file.info:
            self.hic_metadata = cooler_file.info
            # else:
            #     self.hic_metadata = None
            # self.cool_info = deepcopy(cooler_file.info)
            # log.debug('self.hic_metadata {}'.format(self.hic_metadata))
        except Exception as e:  # pylint: disable=W0718
            log.warning("Could not open cooler file. Maybe the path is wrong or the given node is not available.")
            log.warning('The following file was tried to open: %s', self.matrixFileName)
            log.warning("The following nodes are available: %s", cooler.fileops.list_coolers(self.matrixFileName.split("::")[0]))
            return None, e
        if self.chrnameList is None and (self.matrixFileName is None or not self.matrixOnly):
            matrixDataFrame = cooler_file.matrix(balance=False, sparse=True, as_pixels=True)
            used_dtype = np.int32
            if np.iinfo(np.int32).max < cooler_file.info['nbins']:
                used_dtype = np.int64
            count_dtype = matrixDataFrame[0]['count'].dtype
            data = np.empty(cooler_file.info['nnz'], dtype=count_dtype)
            instances = np.empty(cooler_file.info['nnz'], dtype=used_dtype)
            features = np.empty(cooler_file.info['nnz'], dtype=used_dtype)
            i = 0
            size = cooler_file.info['nbins'] // 32
            if size == 0:
                size = 1
            start_pos = 0
            while i < cooler_file.info['nbins']:
                matrixDataFrameChunk = matrixDataFrame[i:i + size]
                _data = matrixDataFrameChunk['count'].values.astype(count_dtype)
                _instances = matrixDataFrameChunk['bin1_id'].values.astype(used_dtype)
                _features = matrixDataFrameChunk['bin2_id'].values.astype(used_dtype)

                data[start_pos:start_pos + len(_data)] = _data
                instances[start_pos:start_pos + len(_instances)] = _instances
                features[start_pos:start_pos + len(_features)] = _features
                start_pos += len(_features)
                i += size
                del _data
                del _instances
                del _features

            if self.matrixFormat is None or self.matrixFormat == 'csr':
                matrix = csr_matrix((data, (instances, features)), shape=(int(cooler_file.info['nbins']), int(cooler_file.info['nbins'])), dtype=count_dtype)
            elif self.matrixFormat == 'lil':
                matrix = lil_matrix((data, (instances, features)), shape=(int(cooler_file.info['nbins']), int(cooler_file.info['nbins'])), dtype=count_dtype)
            elif self.matrixFormat == 'dok':
                matrix = dok_matrix((data, (instances, features)), shape=(int(cooler_file.info['nbins']), int(cooler_file.info['nbins'])), dtype=count_dtype)
            # elif  self.matrixFormat == 'raw':
            #     matrix = [instances, features, data, int(cooler_file.info['nbins'])]
            del data
            del instances
            del features
            gc.collect()
        elif self.chrnameList is None and self.matrixOnly:
            log.debug('Load all at once')
            matrixDataFrame = cooler_file.matrix(balance=False, sparse=True, as_pixels=True)
            used_dtype = np.int64
            # if np.iinfo(np.int32).max < cooler_file.info['nbins']:
            #     used_dtype = np.int64
            count_dtype = matrixDataFrame[0]['count'].dtype
            matrixDataFrameChunk = matrixDataFrame[:]
            data = matrixDataFrameChunk['count'].values.astype(count_dtype)
            instances = matrixDataFrameChunk['bin1_id'].values.astype(used_dtype)
            features = matrixDataFrameChunk['bin2_id'].values.astype(used_dtype)
            # matrix = [_instances, _features, _data, int(cooler_file.info['nbins'])]
            # return matrix, None, None, None, None
        else:
            if len(self.chrnameList) == 1:
                try:
                    if self.distance is None or cooler_file.binsize is None:
                        # load the full chromosome
                        matrix = cooler_file.matrix(balance=False, sparse=True, as_pixels=False).fetch(self.chrnameList[0]).tocsr()  # pylint: disable=E1136
                    else:
                        # load only the values up to a specific distance
                        lo, hi = cooler_file.extent(self.chrnameList[0])  # pylint: disable=E1136
                        dist = self.distance // cooler_file.binsize
                        step = (hi - lo) // 32
                        if step < 1:  # pylint: disable=R1731
                            step = 1
                        mat = lil_matrix((hi - lo, hi - lo), dtype=np.float32)

                        for i0, i1 in cooler.util.partition(lo, hi, step):
                            # fetch stripe
                            pixels = cooler_file.matrix(balance=False, as_pixels=True)[i0:i1, lo:hi]
                            # filter
                            pixels = pixels[(pixels['bin2_id'] - pixels['bin1_id']) < dist]
                            # insert into sparse matrix
                            mat[pixels['bin1_id'] - lo, pixels['bin2_id'] - lo] = pixels['count'].astype(np.float32)
                            del pixels

                        matrix = mat.tocsr()
                        del mat
                        gc.collect()

                except ValueError as ve:
                    log.exception("Wrong chromosome format. Please check UCSC / ensembl notation.")
                    log.exception('Error: %s', str(ve))
            else:
                raise ValueError("Operation to load more as one region is not supported.")

        cut_intervals_data_frame = None
        correction_factors_data_frame = None

        if self.chrnameList is not None:
            if len(self.chrnameList) == 1:
                cut_intervals_data_frame = cooler_file.bins().fetch(self.chrnameList[0])
                log.debug('cut_intervals_data_frame %s', str(list(cut_intervals_data_frame.columns)))
                if self.correctionFactorTable in cut_intervals_data_frame:
                    correction_factors_data_frame = cut_intervals_data_frame[self.correctionFactorTable]
            else:
                raise ValueError("Operation to load more than one chr from bins is not supported.")
        else:
            if self.applyCorrectionLoad and self.correctionFactorTable in cooler_file.bins():
                correction_factors_data_frame = cooler_file.bins()[[self.correctionFactorTable]][:]

            cut_intervals_data_frame = cooler_file.bins()[['chrom', 'start', 'end']][:]

        correction_factors = None
        if correction_factors_data_frame is not None and self.applyCorrectionLoad:
            # apply correction factors to matrix
            # a_i,j = a_i,j * c_i *c_j
            if not self.matrixOnly:
                matrix.eliminate_zeros()
                data = matrix.data
            if len(data) > 1:

                if not self.matrixOnly:
                    matrix.data = matrix.data.astype(float)
                else:
                    data = np.array(data, dtype=float)

                correction_factors = np.array(correction_factors_data_frame.values).flatten()
                # Don't apply correction if weight were just 'nans'
                if np.sum(np.isnan(correction_factors)) != len(correction_factors):
                    # correction_factors = convertNansToZeros(correction_factors)

                    if not self.matrixOnly:
                        # matrix.sort_indices()
                        instances, features = matrix.nonzero()
                    instances_factors = correction_factors[instances]
                    features_factors = correction_factors[features]

                    if self.correctionOperator is None:
                        if self.correctionFactorTable in ['KR', 'VC', 'SQRT_VC']:
                            self.correctionOperator = '/'
                        else:
                            self.correctionOperator = '*'
                        if 'generated-by' in cooler_file.info:
                            log.debug('cooler_file.info[\'generated-by\'] %s %s', cooler_file.info['generated-by'], type(cooler_file.info['generated-by']))
                            generated_by = toString(cooler_file.info['generated-by'])
                            if 'hic2cool' in generated_by:
                                self.hic2cool_version = generated_by.split('-')[1]
                            elif 'hicmatrix' in generated_by:
                                self.hicmatrix_version = generated_by.split('-')[1]

                    instances_factors *= features_factors
                    log.debug('hic2cool: %s', self.hic2cool_version)
                    log.debug('self.correctionOperator: %s', self.correctionOperator)

                    if self.matrixOnly:
                        if self.correctionOperator == '*':
                            log.debug('multi')
                            data *= instances_factors
                        elif self.correctionOperator == '/':
                            log.debug('div')
                            data /= instances_factors
                        log.debug('non')
                        return [instances, features, data, int(cooler_file.info['nbins'])], None, None, None, None

                    if self.correctionOperator == '*':
                        matrix.data *= instances_factors
                        log.debug('foo')
                    elif self.correctionOperator == '/':
                        matrix.data /= instances_factors
                        log.debug('hu')

        elif self.matrixOnly:
            return [instances, features, data, int(cooler_file.info['nbins'])], None, None, None, None

        cut_intervals = []
        if not self.noCutIntervals:
            for values in cut_intervals_data_frame.values:
                cut_intervals.append(tuple([toString(values[0]), values[1], values[2], 1.0]))
        del cut_intervals_data_frame
        del correction_factors_data_frame
        # try to restore nan_bins.
        try:
            # remove possible nan bins introduced by the correction factors
            # to have them part of the nan_bins vector
            mask = np.isnan(matrix.data)
            matrix.data[mask] = 0
            matrix.eliminate_zeros()
            shape = matrix.shape[0] if matrix.shape[0] < matrix.shape[1] else matrix.shape[1]
            nan_bins_indices = np.arange(shape)
            nan_bins_indices = np.setdiff1d(nan_bins_indices, matrix.indices)

            nan_bins = []
            for bin_id in nan_bins_indices:
                if len(matrix[bin_id, :].data) == 0:
                    nan_bins.append(bin_id)
            nan_bins = np.array(nan_bins)
        except Exception:  # pylint: disable=W0718
            nan_bins = None

        distance_counts = None
        # log.debug('self.hic_metadata {}'.format(self.hic_metadata))

        return matrix, cut_intervals, nan_bins, distance_counts, correction_factors

    def create_cooler_input(self, pSymmetric=True, pApplyCorrection=True):
        log.debug('self.hic_metadata 34 %s', self.hic_metadata)

        self.matrix.eliminate_zeros()

        if self.nan_bins is not None and len(self.nan_bins) > 0 and self.fileWasH5:
            # remove nan_bins
            correction_factors = np.ones(self.matrix.shape[0])
            correction_factors[self.nan_bins] = 0
            self.matrix.sort_indices()
            _instances, _features = self.matrix.nonzero()

            instances_factors = correction_factors[_instances]
            features_factors = correction_factors[_features]

            instances_factors = np.logical_not(np.logical_or(instances_factors, features_factors))
            self.matrix.data[instances_factors] = 0
            self.matrix.eliminate_zeros()

        # set possible nans in data to 0
        mask = np.isnan(self.matrix.data)

        self.matrix.data[mask] = 0
        self.matrix.eliminate_zeros()
        # save only the upper triangle of the
        if pSymmetric:
            # symmetric matrix
            self.matrix = triu(self.matrix, format='csr')
        else:
            self.matrix = self.matrix

        self.matrix.eliminate_zeros()

        # create data frame for bins
        # self.cut_intervals is having 4 tuples, bin_data_frame should have 3.correction_factors
        # it looks like it is faster to create it with 4, and drop the last one
        # instead of handling this before.
        bins_data_frame = pd.DataFrame(self.cut_intervals, columns=['chrom', 'start', 'end', 'interactions']).drop('interactions', axis=1)
        dtype_pixel = {'bin1_id': np.int32, 'bin2_id': np.int32, 'count': np.int32}
        log.debug('foo')
        if self.correction_factors is not None and pApplyCorrection:
            dtype_pixel['weight'] = np.float32

            # if the correction was applied by a division, invert it because cool format expects multiplicative if table name is 'weight'
            # https://cooler.readthedocs.io/en/latest/api.html#cooler.Cooler.matrix
            if (self.hic2cool_version is not None and self.hic2cool_version >= '0.5') or self.fileWasH5 or self.correctionOperator == '/':

                log.debug('h5 true')
                self.correction_factors = np.array(self.correction_factors).flatten()
                self.correction_factors = 1 / self.correction_factors
                mask = np.isnan(self.correction_factors)
                self.correction_factors[mask] = 0
                mask = np.isinf(self.correction_factors)
                self.correction_factors[mask] = 0
                self.correctionOperator = '*'
                log.debug('inverted correction factors')
            weight = convertNansToOnes(np.array(self.correction_factors).flatten())
            log.debug('weight %s', weight)
            bins_data_frame = bins_data_frame.assign(weight=weight)

            log.debug("Reverting correction factors on matrix...")
            instances, features = self.matrix.nonzero()
            self.correction_factors = np.array(self.correction_factors)

            # do not apply if correction factors are just 1's
            instances_factors = self.correction_factors[instances]
            features_factors = self.correction_factors[features]

            instances_factors *= features_factors

            self.matrix.data = self.matrix.data.astype(float)

            # Apply the invert operation to get the original data
            if self.correctionOperator == '*' or self.correctionOperator is None:
                self.matrix.data /= instances_factors

            instances_factors = None
            features_factors = None

            self.matrix.eliminate_zeros()

        if self.correction_factors is not None and pApplyCorrection is False:
            dtype_pixel['weight'] = np.float32
            weight = convertNansToOnes(np.array(self.correction_factors).flatten())
            bins_data_frame = bins_data_frame.assign(weight=weight)
            log.debug('weight 2: %s', weight)
        instances, features = self.matrix.nonzero()

        matrix_data_frame = pd.DataFrame(instances, columns=['bin1_id'], dtype=np.int32)
        del instances
        matrix_data_frame = matrix_data_frame.assign(bin2_id=features)
        del features

        if self.enforceInteger:
            dtype_pixel['count'] = np.int32
            data = np.rint(self.matrix.data)
            matrix_data_frame = matrix_data_frame.assign(count=data)
        else:
            matrix_data_frame = matrix_data_frame.assign(count=self.matrix.data)

        if not self.enforceInteger and self.matrix.dtype not in [np.int32, int]:
            log.debug("Writing non-standard cooler matrix. Datatype of matrix['count'] is: %s", self.matrix.dtype)
            dtype_pixel['count'] = self.matrix.dtype
        split_factor = 1
        if len(self.matrix.data) > 1e7:
            split_factor = 1e4
            matrix_data_frame = np.array_split(matrix_data_frame, split_factor)

        if self.appendData:
            self.appendData = 'a'
        else:
            self.appendData = 'w'

        info = {}
        # these fields are created by cooler lib. Can cause errors if not deleted.
        if 'metadata' in info:
            if self.hic_metadata is None:
                self.hic_metadata = info['metadata']
            del info['metadata']
        if 'bin-size' in info:
            del info['bin-size']
        if 'bin-type' in info:
            del info['bin-type']

        info['format'] = str('HDF5::Cooler')
        info['format-url'] = str('https://github.com/mirnylab/cooler')
        info['generated-by'] = str('HiCMatrix-' + version('HiCMatrix'))
        info['generated-by-cooler-lib'] = str('cooler-' + version('cooler'))

        info['tool-url'] = str('https://github.com/deeptools/HiCMatrix')

        if self.hic_metadata is not None and 'matrix-generated-by' in self.hic_metadata:
            info['matrix-generated-by'] = str(self.hic_metadata['matrix-generated-by'])
            del self.hic_metadata['matrix-generated-by']
        if self.hic_metadata is not None and 'matrix-generated-by-url' in self.hic_metadata:
            info['matrix-generated-by-url'] = str(self.hic_metadata['matrix-generated-by-url'])
            del self.hic_metadata['matrix-generated-by-url']
        log.debug('self.hic_metadata %s', self.hic_metadata)
        if self.hic_metadata is not None and 'genome-assembly' in self.hic_metadata:
            info['genome-assembly'] = str(self.hic_metadata['genome-assembly'])
            del self.hic_metadata['genome-assembly']

        return bins_data_frame, matrix_data_frame, dtype_pixel, info

    def save(self, pFileName, pSymmetric=True, pApplyCorrection=True):
        log.debug('Save in cool format11112323')

        bins_data_frame, matrix_data_frame, dtype_pixel, info = self.create_cooler_input(pSymmetric=pSymmetric, pApplyCorrection=pApplyCorrection)
        local_temp_dir = os.path.dirname(os.path.realpath(pFileName))
        cooler.create_cooler(cool_uri=pFileName,
                             bins=bins_data_frame,
                             pixels=matrix_data_frame,
                             mode=self.appendData,
                             dtypes=dtype_pixel,
                             ordered=True,
                             metadata=info,

                             temp_dir=local_temp_dir)

        log.debug('info %s', info)
        if self.appendData == 'w':
            fileName = pFileName.split('::')[0]
            with h5py.File(fileName, 'r+') as h5file:
                h5file.attrs.update(info)
                h5file.close()



import logging
import os.path
import sys
import warnings
# from past.builtins import zip
from collections import OrderedDict
from os import unlink
from tempfile import NamedTemporaryFile

import numpy as np
import numpy.testing as nt
import pytest
from intervaltree import Interval, IntervalTree
from scipy.sparse import coo_matrix, csr_matrix
from six import iteritems

from hicmatrix import HiCMatrix as hm

log = logging.getLogger(__name__)

warnings.filterwarnings("ignore")

ROOT = os.path.join(os.path.dirname(os.path.abspath(__file__)), "test_data/")


def test_load_h5_save_and_load_cool():
    hic = hm.hiCMatrix(ROOT + 'Li_et_al_2015.h5')

    outfile = NamedTemporaryFile(suffix='.cool', prefix='hicexplorer_test')  # pylint: disable=R1732
    hic.matrixFileHandler = None
    hic.save(pMatrixName=outfile.name)

    hic_cool = hm.hiCMatrix(outfile.name)

    nt.assert_equal(hic_cool.matrix.data, hic.matrix.data)
    chrom_cool, start_cool, end_cool, _ = list(zip(*hic_cool.cut_intervals))
    chrom, start, end, _ = list(zip(*hic_cool.cut_intervals))

    nt.assert_equal(chrom_cool, chrom)
    nt.assert_equal(start_cool, start)
    nt.assert_equal(end_cool, end)


def test_load_h5_load_cool_weight():
    hic_h5 = hm.hiCMatrix(ROOT + 'Li_et_al_2015.h5')
    hic_cool = hm.hiCMatrix(ROOT + 'Li_et_al_2015.cool')

    # there is always a small gap due to rounding errors and inaccurate floating operations
    # test if it is equal for up to 10 decimal positions
    nt.assert_almost_equal(hic_cool.matrix.data, hic_h5.matrix.data, decimal=10)
    chrom_cool, start_cool, end_cool, _ = list(zip(*hic_cool.cut_intervals))
    chrom, start, end, _ = list(zip(*hic_cool.cut_intervals))

    nt.assert_equal(chrom_cool, chrom)
    nt.assert_equal(start_cool, start)
    nt.assert_equal(end_cool, end)


def test_load_h5_save_and_load_cool_2():
    hic = hm.hiCMatrix(ROOT + 'small_test_matrix.h5')

    outfile = NamedTemporaryFile(suffix='.cool', prefix='hicexplorer_test')  # pylint: disable=R1732
    hic.matrixFileHandler = None
    hic.save(pMatrixName=outfile.name)

    hic_cool = hm.hiCMatrix(outfile.name)

    nt.assert_equal(hic_cool.matrix.data, hic.matrix.data)
    chrom_cool, start_cool, end_cool, _ = list(zip(*hic_cool.cut_intervals))
    chrom, start, end, _ = list(zip(*hic_cool.cut_intervals))

    nt.assert_equal(chrom_cool, chrom)
    nt.assert_equal(start_cool, start)
    nt.assert_equal(end_cool, end)


def test_load_cool_save_and_load_h5():
    hic = hm.hiCMatrix(ROOT + 'Li_et_al_2015.cool')

    outfile = NamedTemporaryFile(suffix='.h5', prefix='hicexplorer_test')  # pylint: disable=R1732
    hic.matrixFileHandler = None
    hic.save(pMatrixName=outfile.name)

    hic_cool = hm.hiCMatrix(outfile.name)

    nt.assert_equal(hic_cool.matrix.data, hic.matrix.data)
    chrom_cool, start_cool, end_cool, _ = list(zip(*hic_cool.cut_intervals))
    chrom, start, end, _ = list(zip(*hic_cool.cut_intervals))

    nt.assert_equal(chrom_cool, chrom)
    nt.assert_equal(start_cool, start)
    nt.assert_equal(end_cool, end)


def test_save_load_cool():
    outfile = '/tmp/matrix.cool'
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('a', 30, 40, 1), ('b', 40, 50, 1)]
    hic = hm.hiCMatrix()
    hic.nan_bins = []
    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    # make matrix symmetric
    hic.setMatrix(hic.matrix, cut_intervals)
    hic.fillLowerTriangle()
    # hic.correction_factors = np.array([0.5, 1, 2, 3, 4])
    # hic.nan_bins = np.array([4])

    hic.save(outfile)

    cool_obj = hm.hiCMatrix(outfile)
    # nt.assert_equal(hic.correction_factors, cool_obj.correction_factors)
    nt.assert_equal(hic.matrix.data, cool_obj.matrix.data)
    nt.assert_equal(hic.matrix.indices, cool_obj.matrix.indices)
    nt.assert_equal(hic.matrix.indptr, cool_obj.matrix.indptr)
    nt.assert_equal(hic.nan_bins, cool_obj.nan_bins)

    nt.assert_equal(hic.cut_intervals, cool_obj.cut_intervals)
    unlink(outfile)


def test_save_load_h5():
    outfile = '/tmp/matrix.h5'
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('a', 30, 40, 1), ('b', 40, 50, 1)]
    hic = hm.hiCMatrix()
    hic.nan_bins = []
    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    # make matrix symmetric
    hic.setMatrix(hic.matrix, cut_intervals)
    hic.fillLowerTriangle()
    # hic.correction_factors = np.array([0.5, 1, 2, 3, 4])
    # hic.nan_bins = np.array([4])

    hic.save(outfile)

    h5_obj = hm.hiCMatrix(outfile)
    # nt.assert_equal(hic.correction_factors, h5_obj.correction_factors)
    nt.assert_equal(hic.matrix.data, h5_obj.matrix.data)
    nt.assert_equal(hic.matrix.indices, h5_obj.matrix.indices)
    nt.assert_equal(hic.matrix.indptr, h5_obj.matrix.indptr)
    nt.assert_equal(hic.nan_bins, h5_obj.nan_bins)

    nt.assert_equal(hic.cut_intervals, h5_obj.cut_intervals)
    unlink(outfile)


@pytest.mark.xfail
def test_save_load_other_formats_fail():
    pMatrixFile = ROOT + 'test_matrix.hicpro'
    # pBedFileHicPro = ROOT + 'test_matrix.bed'  # no parameter for this in hiCMatrix::__init__() anyway
    # hic_matrix = hm.hiCMatrix(pMatrixFile=pMatrixFile)
    # out, err = capsys.readouterr()
    # assert out == 'matrix file not given'
    pMatrixFile = ROOT + 'test_matrix.homer'
    hm.hiCMatrix(pMatrixFile=pMatrixFile)


def test_convert_to_zscore_matrix():

    # make test matrix
    m_size = 100
    mat = np.triu(np.random.randint(0, 101, (m_size, m_size)))
    # add a number of zeros
    mat[mat < 90] = 0
    # import ipdb;ipdb.set_trace()
    mu = dict([(idx, mat.diagonal(idx).mean()) for idx in range(mat.shape[0])])  # pylint: disable=R1717
    std = dict([(idx, np.std(mat.diagonal(idx)))  # pylint: disable=R1717
                for idx in range(mat.shape[0])])

    # compute z-score for test matrix
    zscore_mat = np.zeros((m_size, m_size))
    for _i in range(mat.shape[0]):
        for _j in range(mat.shape[0]):
            if _j >= _i:
                diag = _j - _i
                if std[diag] == 0:
                    zscore = np.nan
                else:
                    zscore = (mat[_i, _j] - mu[diag]) / std[diag]
                zscore_mat[_i, _j] = zscore

    # make Hi-C matrix based on test matrix
    hic = hm.hiCMatrix()
    hic.matrix = csr_matrix(mat)
    cut_intervals = [('chr', idx, idx + 10, 0) for idx in range(0, mat.shape[0] * 10, 10)]
    hic.setMatrix(hic.matrix, cut_intervals)
    hic.convert_to_zscore_matrix()

    nt.assert_almost_equal(hic.matrix.todense(), zscore_mat)


def test_convert_to_zscore_matrix_2():

    # load test matrix
    hic = hm.hiCMatrix(ROOT + 'Li_et_al_2015.h5')
    hic.maskBins(hic.nan_bins)

    mat = hic.matrix.todense()
    max_depth = 10000
    bin_size = hic.getBinSize()
    max_depth_in_bins = int(float(max_depth) / bin_size)

    m_size = mat.shape[0]
    # compute matrix values per distance
    _, start, _, _ = list(zip(
        *hm.hiCMatrix.fit_cut_intervals(hic.cut_intervals)))
    dist_values = {}
    sys.stderr.write("Computing values per distance for each matrix entry\n")

    for _i in range(mat.shape[0]):
        for _j in range(mat.shape[0]):
            if _j >= _i:
                # dist is translated to bins
                dist = int(float(start[_j] - start[_i]) / bin_size)
                if dist <= max_depth_in_bins:
                    if dist not in dist_values:
                        dist_values[dist] = []
                    dist_values[dist].append(mat[_i, _j])

    mu = {}
    std = {}
    for dist, values in iteritems(dist_values):
        mu[dist] = np.mean(values)
        std[dist] = np.std(values)

    # compute z-score for test matrix
    sys.stderr.write("Computing zscore for each matrix entry\n")
    zscore_mat = np.full((m_size, m_size), np.nan)
    for _i in range(mat.shape[0]):
        for _j in range(mat.shape[0]):
            if _j >= _i:
                dist = int(float(start[_j] - start[_i]) / bin_size)
                if dist <= max_depth_in_bins:
                    zscore = (mat[_i, _j] - mu[dist]) / std[dist]
                    zscore_mat[_i, _j] = zscore

    # compare with zscore from class
    hic.convert_to_zscore_matrix(maxdepth=max_depth)

    # from numpy.testing import assert_almost_equal
    # only the main diagonal is check. Other diagonals show minimal differences
    nt.assert_almost_equal(hic.matrix.todense().diagonal(
        0).A1, zscore_mat.diagonal(0))


def test_dist_list_to_dict():
    hic = hm.hiCMatrix()

    data = np.array([1, 8, 5, 3, 0, 4, 15, 5, 1, 0, 0, 2, 0, 1, 0])
    dist_list = np.array(
        [0, 10, 20, 30, -1, 0, 10, 20, -1, 0, 10, -1, 0, -1, 0])

    distance = hic.dist_list_to_dict(data, dist_list)

    nt.assert_equal(distance[-1], [0, 1, 2, 1])
    nt.assert_equal(distance[0], [1, 4, 0, 0, 0])
    nt.assert_equal(distance[10], [8, 15, 0])
    nt.assert_equal(distance[20], [5, 5])
    nt.assert_equal(distance[30], [3])

    data = np.array([0, 100, 200, 0, 100, 200, 0, 100, 0])
    dist_list = np.array([0, 100, 200, 0, 100, 200, 0, 100, 0])

    distance = hic.dist_list_to_dict(data, dist_list)

    nt.assert_equal(distance[0], [0, 0, 0, 0])
    nt.assert_equal(distance[100], [100, 100, 100])
    nt.assert_equal(distance[200], [200, 200])


def test_keepOnlyTheseChr():
    chromosome_list = ['chrX', 'chr2RHet']

    hic = hm.hiCMatrix(ROOT + 'small_test_matrix.h5')

    hic.keepOnlyTheseChr(chromosome_list)

    nt.assert_equal(hic.getChrNames().sort(), chromosome_list.sort())


def test_save():
    """
    Test will not cover testing of following formats due to unsupported file_formats (see __init__ of class hiCMatrix):

    * ren
    * lieberman
    * GInteractions

    see also single test for these formats (marked as xfail)
    """

    outfile_cool = NamedTemporaryFile(suffix='.cool', delete=False)  # pylint: disable=R1732
    outfile_cool.close()

    outfile_h5 = NamedTemporaryFile(suffix='.h5', delete=False)  # pylint: disable=R1732
    outfile_h5.close()

    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('a', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)
    hic.fillLowerTriangle()

    # test .h5
    hic.save(outfile_h5.name)
    h5_test = hm.hiCMatrix(outfile_h5.name)

    # test cool
    hic.matrixFileHandler = None
    hic.save(outfile_cool.name)
    cool_test = hm.hiCMatrix(outfile_cool.name)

    nt.assert_equal(hic.getMatrix(), h5_test.getMatrix())
    nt.assert_equal(hic.getMatrix(), cool_test.getMatrix())


def test_diagflat():
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('a', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)
    hic.fillLowerTriangle()

    hic.diagflat(value=1000)
    nt.assert_equal(
        np.array([1000 for x in range(matrix.shape[0])]), hic.matrix.diagonal())

    hic.diagflat()
    nt.assert_equal(
        np.array([np.nan for x in range(5)]), hic.matrix.diagonal())


def test_filterOutInterChrCounts():
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)
    hic.fillLowerTriangle()
    hic.filterOutInterChrCounts()

    filtered_matrix = np.array([[1, 8, 5, 0, 0],
                                [8, 4, 15, 0, 0],
                                [5, 15, 0, 0, 0],
                                [0, 0, 0, 0, 1],
                                [0, 0, 0, 1, 0]])

    nt.assert_equal(hic.getMatrix(), filtered_matrix)

    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]
    hic = hm.hiCMatrix()
    hic.nan_bins = []
    matrix = np.array([[0, 10, 5, 3, 0],
                       [0, 0, 15, 5, 1],
                       [0, 0, 0, 7, 3],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    # make the matrix symmetric:
    hic.matrix = csr_matrix(matrix + matrix.T)
    hic.setMatrix(csr_matrix(matrix + matrix.T, dtype=np.int32), cut_intervals)

    filtered = hic.filterOutInterChrCounts().todense()
    test_matrix = np.array([[0, 10, 5, 0, 0],
                            [10, 0, 15, 0, 0],
                            [5, 15, 0, 0, 0],
                            [0, 0, 0, 0, 1],
                            [0, 0, 0, 1, 0]], dtype='i4')

    nt.assert_equal(filtered, test_matrix)


def test_setMatrixValues_success():
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    new_matrix = np.array([[10, 80, 50, 30, 0],
                           [0, 40, 150, 50, 10],
                           [0, 0, 0, 0, 20],
                           [0, 0, 0, 0, 10],
                           [0, 0, 0, 0, 0]])

    hic.setMatrixValues(new_matrix)

    nt.assert_equal(hic.getMatrix(), new_matrix)


def test_setMatrixValues_fail():
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1)]

    new_matrix = np.array([[10, 80, 50, 30],
                           [0, 40, 150, 50],
                           [0, 0, 0, 0],
                           [0, 0, 0, 0]])
    with pytest.raises(AssertionError):
        hic.setMatrixValues(new_matrix)


def test_setCorrectionFactors_success():
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    assert hic.correction_factors is None

    hic.setCorrectionFactors([5, 5, 5, 5, 5])

    nt.assert_equal(hic.correction_factors, [5, 5, 5, 5, 5])


def test_setCorrectionFactors_fail():
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    assert hic.correction_factors is None
    with pytest.raises(AssertionError):
        hic.setCorrectionFactors([5, 5, 5, 5])


def test_reorderChromosomes():
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    new_chr_order = ['b', 'a']
    hic.reorderChromosomes(new_chr_order)

    nt.assert_equal(hic.chrBinBoundaries, OrderedDict(
        [('b', (0, 2)), ('a', (2, 5))]))

    old_chr_order = ['a', 'b']
    hic.reorderChromosomes(old_chr_order)

    nt.assert_equal(hic.chrBinBoundaries, OrderedDict(
        [('a', (0, 3)), ('b', (3, 5))]))


def test_reorderChromosomes_fail():
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    # name 'c' not in chromosome names, thus fail
    false_chr_order = ['a', 'b', 'c']
    with pytest.raises(Exception) as context:
        hic.reorderChromosomes(false_chr_order)
    assert "Chromosome name 'c' not found." in str(context.value)


def test_reorderBins():
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    nt.assert_equal(hic.getMatrix(), matrix)

    new_order = [0, 1, 3, 2, 4]
    new_matrix = np.array([[1, 8, 3, 5, 0],
                           [0, 4, 5, 15, 1],
                           [0, 0, 0, 0, 1],
                           [0, 0, 0, 0, 2],
                           [0, 0, 0, 0, 0]])

    hic.reorderBins(new_order)

    nt.assert_equal(hic.getMatrix(), new_matrix)

    hic.reorderBins(new_order)

    nt.assert_equal(hic.getMatrix(), matrix)

    # order smaller than original matrix should delete unused ids
    small_order = [2, 3]
    small_matrix = np.array([[0, 0],
                             [0, 0]])

    hic.reorderBins(small_order)

    nt.assert_equal(hic.getMatrix(), small_matrix)
    nt.assert_equal(hic.matrix.shape, small_matrix.shape)
    nt.assert_equal(hic.chrBinBoundaries, OrderedDict(
        [('a', (0, 1)), ('b', (1, 2))]))
    nt.assert_equal(hic.cut_intervals, [('a', 20, 30, 1), ('b', 30, 40, 1)])
    nt.assert_equal(hic.nan_bins, [])


def test_maskBins():
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    nt.assert_equal(hic.getMatrix(), matrix)
    nt.assert_equal(hic.orig_bin_ids, [])

    new_matrix = np.array([[0, 0, 2],
                           [0, 0, 1],
                           [0, 0, 0]])

    masking_ids = [0, 1]
    hic.maskBins(masking_ids)

    nt.assert_equal(hic.getMatrix(), new_matrix)
    nt.assert_equal(sorted(hic.orig_cut_intervals), sorted([('a', 0, 10, 1), ('a', 10, 20, 1),
                                                            ('a', 20, 30,
                                                             1), ('b', 30, 40, 1),
                                                            ('b', 40, 50, 1)]))
    nt.assert_equal(sorted(hic.cut_intervals), sorted([('a', 20, 30, 1), ('b', 30, 40, 1),
                                                       ('b', 40, 50, 1)]))
    nt.assert_equal(hic.chrBinBoundaries, OrderedDict(
        [('a', (0, 1)), ('b', (1, 3))]))
    nt.assert_equal(sorted(hic.orig_bin_ids), sorted([0, 1, 2, 3, 4]))

    # direct return if masking_ids is None or has len() == 0, thus no changes to matrix
    masking_ids = None
    hic.maskBins(masking_ids)

    nt.assert_equal(hic.getMatrix(), new_matrix)
    nt.assert_equal(sorted(hic.orig_cut_intervals), sorted([('a', 0, 10, 1), ('a', 10, 20, 1),
                                                            ('a', 20, 30,
                                                             1), ('b', 30, 40, 1),
                                                            ('b', 40, 50, 1)]))
    nt.assert_equal(sorted(hic.cut_intervals), sorted([('a', 20, 30, 1), ('b', 30, 40, 1),
                                                       ('b', 40, 50, 1)]))
    nt.assert_equal(hic.chrBinBoundaries, OrderedDict(
        [('a', (0, 1)), ('b', (1, 3))]))

    masking_ids = []

    hic.maskBins(masking_ids)

    nt.assert_equal(hic.getMatrix(), new_matrix)
    nt.assert_equal(sorted(hic.orig_cut_intervals), sorted([('a', 0, 10, 1), ('a', 10, 20, 1),
                                                            ('a', 20, 30,
                                                             1), ('b', 30, 40, 1),
                                                            ('b', 40, 50, 1)]))
    nt.assert_equal(sorted(hic.cut_intervals), sorted([('a', 20, 30, 1), ('b', 30, 40, 1),
                                                       ('b', 40, 50, 1)]))
    nt.assert_equal(hic.chrBinBoundaries, OrderedDict(
        [('a', (0, 1)), ('b', (1, 3))]))

    nt.assert_equal(sorted(hic.orig_bin_ids), sorted([0, 1, 2, 3, 4]))


def test_update_matrix():
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    nt.assert_equal(hic.getMatrix(), matrix)

    new_cut_intervals = [('c', 0, 10, 1), ('d', 10, 20, 1), ('d', 20, 30, 1)]

    new_matrix = np.array([[3, 6, 4],
                           [np.nan, 0, 2],
                           [1, 0, 0]])
    try:
        hic.update_matrix(new_matrix, new_cut_intervals)
    except AttributeError:
        pass
    # if matrix.shape[0] not equal to length of cut_intervals assertionError is raised
    short_cut_intervals = [('c', 0, 10, 1), ('d', 10, 20, 1)]

    with pytest.raises(AssertionError):
        hic.update_matrix(new_matrix, short_cut_intervals)

    # if matrix contains masked bins exception is raised
    masking_ids = [0, 1]
    hic.maskBins(masking_ids)

    with pytest.raises(Exception):
        hic.update_matrix(new_matrix, new_cut_intervals)


def test_restoreMaskedBins():
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    nt.assert_equal(hic.getMatrix(), matrix)
    nt.assert_equal(hic.orig_bin_ids, [])

    # function should directly return if there are no masked_bins
    hic.restoreMaskedBins()

    nt.assert_equal(hic.getMatrix(), matrix)
    nt.assert_equal(hic.orig_bin_ids, [])

    # test general use
    # first get some masked bins
    masking_ids = [0, 1]
    hic.maskBins(masking_ids)

    new_matrix = np.array([[0, 0, 2],
                           [0, 0, 1],
                           [0, 0, 0]])

    nt.assert_equal(hic.getMatrix(), new_matrix)
    nt.assert_equal(sorted(hic.orig_bin_ids), sorted([0, 1, 2, 3, 4]))

    # and now restore masked bins
    hic.restoreMaskedBins()

    result_matrix = np.array([[np.nan, np.nan, np.nan, np.nan, np.nan],
                              [np.nan, np.nan, np.nan, np.nan, np.nan],
                              [np.nan, np.nan, 0, 0, 2],
                              [np.nan, np.nan, 0, 0, 1],
                              [np.nan, np.nan, 0, 0, 0]])

    nt.assert_equal(hic.getMatrix(), result_matrix)
    nt.assert_equal(hic.orig_bin_ids, [])

    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('a', 30, 40, 1), ('b', 40, 50, 1)]
    hic = hm.hiCMatrix()
    hic.nan_bins = []
    matrix = np.array([[0, 10, 5, 3, 0],
                       [0, 0, 15, 5, 1],
                       [0, 0, 0, 7, 3],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]], dtype=np.int32)

    # make the matrix symmetric:
    hic.matrix = csr_matrix(matrix + matrix.T)
    hic.setMatrix(csr_matrix(matrix + matrix.T), cut_intervals)

    # Add masked bins masked bins
    hic.maskBins([3])

    matrix = hic.matrix.todense()
    test_matrix = np.array([[0, 10, 5, 0],
                            [10, 0, 15, 1],
                            [5, 15, 0, 3],
                            [0, 1, 3, 0]], dtype=np.int32)

    nt.assert_equal(matrix, test_matrix)

    cut_int = hic.cut_intervals
    test_cut_int = [('a', 0, 10, 1), ('a', 10, 20, 1), ('a', 20, 30, 1), ('b', 40, 50, 1)]

    nt.assert_equal(cut_int, test_cut_int)

    hic.restoreMaskedBins()

    dense = hic.matrix.todense()
    test_dense = np.array([[0., 10., 5., 0., 0.],
                           [10., 0., 15., 0., 1.],
                           [5., 15., 0., 0., 3.],
                           [0., 0., 0., 0., 0.],
                           [0., 1., 3., 0., 0.]])

    nt.assert_equal(dense, test_dense)

    cut_int = hic.cut_intervals
    test_cut_int = [('a', 0, 10, 1), ('a', 10, 20, 1), ('a', 20, 30, 1),
                    ('a', 30, 40, 1), ('b', 40, 50, 1)]

    nt.assert_equal(cut_int, test_cut_int)


def test_reorderMatrix():
    orig = (1, 3)
    dest = 2

    # get matrix
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    nt.assert_equal(hic.getMatrix(), matrix)

    # reorder matrix
    hic.reorderMatrix(orig, dest)

    new_matrix = np.array([[1, 3, 8, 5, 0],
                           [0, 0, 0, 0, 1],
                           [0, 5, 4, 15, 1],
                           [0, 0, 0, 0, 2],
                           [0, 0, 0, 0, 0]])

    new_cut_intervals = [('a', 0, 10, 1), ('b', 30, 40, 1),
                         ('a', 10, 20, 1), ('a', 20, 30, 1), ('b', 40, 50, 1)]

    # check if it is equal
    nt.assert_equal(hic.getMatrix(), new_matrix)
    nt.assert_equal(hic.matrix.shape, new_matrix.shape)
    nt.assert_equal(hic.cut_intervals, new_cut_intervals)


def test_truncTrans():
    # get matrix
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[-1, 8, 5, 3, 0],
                       [np.nan, 4, 15, 5, 100],
                       [0, 0, 0, 0, 2000],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    nt.assert_equal(hic.getMatrix(), matrix)

    # define expected outcome
    new_matrix = np.array([[-1., 8., 5., 3., 0.],
                           [np.nan, 4., 15., 5., 1.e+2],
                           [0., 0., 0., 0., 2.e+3],
                           [0., 0., 0., 0., 1.],
                           [0., 0., 0., 0., 0.]])

    # truncTrans of matrix
    hic.truncTrans()

    # test against expected outcome
    nt.assert_equal(hic.getMatrix(), new_matrix)

    # reset matrix
    matrix = np.array([[-1, 8, 5, 3, 0],
                       [np.nan, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])
    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    # method should directly return if nothing to do, matrix stays the same
    hic.truncTrans()
    nt.assert_equal(hic.getMatrix(), matrix)


def test_printchrtoremove(capsys):
    # get matrix
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    nt.assert_equal(hic.getMatrix(), matrix)

    # first test exception message for no self.prev_to_remove
    to_remove = [0, 1]

    with pytest.raises(Exception):
        hic.printchrtoremove(to_remove)

        captured = capsys.readouterr()
        assert captured.out == "No self.prev_to_remove defined, defining it now."

        nt.assert_equal(hic.prev_to_remove, np.array(to_remove))

    nt.assert_equal(hic.orig_bin_ids, [])

    # also test with masked_bins
    hic.maskBins(to_remove)

    assert len(hic.orig_bin_ids) > 0

    hic.printchrtoremove(to_remove)

    nt.assert_equal(hic.prev_to_remove, np.array(to_remove))


def test_get_chromosome_sizes_real():
    # get matrix
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    nt.assert_equal(hic.getMatrix(), matrix)

    # define expected outcome
    expected_sizes = OrderedDict([('a', 31), ('b', 21)])

    chrom_sizes = hic.get_chromosome_sizes_real()

    nt.assert_equal(chrom_sizes, expected_sizes)

    # define new intervals and test again
    new_cut_intervals = [('a', 0, 10, 1), ('b', 10, 20, 1),
                         ('b', 20, 30, 1), ('c', 30, 40, 1), ('c', 40, 90, 1)]

    expected_sizes = OrderedDict([('a', 11), ('b', 21), ('c', 61)])

    hic.setMatrix(hic.matrix, new_cut_intervals)

    chrom_sizes = hic.get_chromosome_sizes_real()

    nt.assert_equal(chrom_sizes, expected_sizes)


def test_get_chromosome_sizes():
    # get matrix
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    nt.assert_equal(hic.getMatrix(), matrix)

    # define expected outcome
    expected_sizes = OrderedDict([('a', 30), ('b', 50)])

    chrom_sizes = hic.get_chromosome_sizes()

    nt.assert_equal(chrom_sizes, expected_sizes)

    # define new intervals and test again
    new_cut_intervals = [('a', 0, 10, 1), ('b', 10, 20, 1),
                         ('b', 20, 30, 1), ('c', 30, 40, 1), ('c', 40, 90, 1)]

    expected_sizes = OrderedDict([('a', 10), ('b', 30), ('c', 90)])

    hic.setMatrix(hic.matrix, new_cut_intervals)

    chrom_sizes = hic.get_chromosome_sizes()

    nt.assert_equal(chrom_sizes, expected_sizes)


def test_intervalListToIntervalTree(capsys):
    # get matrix
    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    nt.assert_equal(hic.getMatrix(), matrix)

    # empty list should raise AssertionError
    interval_list = []
    with pytest.raises(AssertionError):
        hic.intervalListToIntervalTree(interval_list)

        captured = capsys.readouterr()
        assert captured.out == "Interval list is empty"

    # test with correct interval_list
    interval_list = [('a', 0, 10, 1), ('a', 10, 20, 1), ('b', 20, 30, 1), ('b', 30, 50, 1),
                     ('b', 50, 100, 1), ('c', 100, 200, 1), ('c', 200, 210, 1),
                     ('d', 210, 220, 1), ('e', 220, 250)]

    tree, boundaries = hic.intervalListToIntervalTree(interval_list)

    # test tree
    nt.assert_equal(tree['a'], IntervalTree([Interval(0, 10, 0), Interval(10, 20, 1)]))
    nt.assert_equal(tree['b'], IntervalTree([Interval(20, 30, 2), Interval(30, 50, 3),
                                             Interval(50, 100, 4)]))
    nt.assert_equal(tree['c'], IntervalTree([Interval(100, 200, 5), Interval(200, 210, 6)]))
    nt.assert_equal(tree['d'], IntervalTree([Interval(210, 220, 7)]))
    nt.assert_equal(tree['e'], IntervalTree([Interval(220, 250, 8)]))

    # test boundaries
    nt.assert_equal(boundaries, OrderedDict([('a', (0, 2)), ('b', (2, 5)), ('c', (5, 7)),
                                             ('d', (7, 8)), ('e', (8, 9))]))


def test_fillLowerTriangle():
    A = csr_matrix(np.array([[12, 5, 3, 2, 0], [0, 11, 4, 1, 1], [0, 0, 9, 6, 0],
                             [0, 0, 0, 10, 0], [0, 0, 0, 0, 0]]), dtype=np.int32)
    hic = hm.hiCMatrix()
    hic.matrix = A
    hic.fillLowerTriangle()
    B = hic.matrix
    test_matrix = np.array([[12, 5, 3, 2, 0],
                            [5, 11, 4, 1, 1],
                            [3, 4, 9, 6, 0],
                            [2, 1, 6, 10, 0],
                            [0, 1, 0, 0, 0]], dtype='i4')

    nt.assert_equal(B.todense(), test_matrix)


def test_getDistList():
    row, col = np.triu_indices(5)
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('a', 30, 40, 1), ('b', 40, 50, 1)]
    dist_list, chrom_list = hm.hiCMatrix.getDistList(row, col, cut_intervals)

    matrix = coo_matrix((dist_list, (row, col)), shape=(5, 5), dtype=np.int32).todense()
    test_matrix = np.array([[0, 10, 20, 30, -1],
                            [0, 0, 10, 20, -1],
                            [0, 0, 0, 10, -1],
                            [0, 0, 0, 0, -1],
                            [0, 0, 0, 0, 0]], dtype='i4')
    nt.assert_equal(matrix, test_matrix)

    chrom_list = chrom_list.tolist()
    test_chrom_list = ['a', 'a', 'a', 'a', '', 'a', 'a', 'a', '', 'a', 'a', '', 'a',
                       '', 'b']

    nt.assert_equal(chrom_list, test_chrom_list)


def test_convert_to_obs_exp_matrix():
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('a', 30, 40, 1), ('b', 40, 50, 1)]
    hic = hm.hiCMatrix()
    hic.nan_bins = []
    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 7, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    obs_exp_matrix = hic.convert_to_obs_exp_matrix().todense()
    test_matrix = np.array([[1., 0.8, 1., 1., 0.],
                            [0., 4., 1.5, 1., 1.],
                            [0., 0., 0., 0.7, 2.],
                            [0., 0., 0., 0., 1.],
                            [0., 0., 0., 0., 0.]])

    nt.assert_almost_equal(obs_exp_matrix, test_matrix)

    hic.matrix = csr_matrix(matrix)
    obs_exp_matrix = hic.convert_to_obs_exp_matrix(maxdepth=20).todense()
    test_matrix = np.array([[1., 0.8, 1., 0., 0.],
                            [0., 4., 1.5, 1., 0.],
                            [0., 0., 0., 0.7, np.nan],
                            [0., 0., 0., 0., np.nan],
                            [0., 0., 0., 0., 0.]])

    nt.assert_almost_equal(obs_exp_matrix, test_matrix)

    hic.matrix = csr_matrix(matrix)

    obs_exp_matrix = hic.convert_to_obs_exp_matrix(zscore=True).todense()
    test_matrix = np.array([[0., -0.56195149, np.nan, np.nan, -1.41421356],
                            [0., 1.93649167, 1.40487872, np.nan, 0.],
                            [0., 0., -0.64549722, -0.84292723, 1.41421356],
                            [0., 0., 0., -0.64549722, 0.],
                            [0., 0., 0., 0., -0.64549722]])

    nt.assert_almost_equal(obs_exp_matrix, test_matrix)


def test_maskChromosomes():

    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)
    hic.maskChromosomes(['a'])


@pytest.mark.xfail
def test_maskChromosomes_fail():

    hic = hm.hiCMatrix()
    cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
                     ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]

    hic.nan_bins = []

    matrix = np.array([[1, 8, 5, 3, 0],
                       [0, 4, 15, 5, 1],
                       [0, 0, 0, 0, 2],
                       [0, 0, 0, 0, 1],
                       [0, 0, 0, 0, 0]])

    hic.matrix = csr_matrix(matrix)
    hic.setMatrix(hic.matrix, cut_intervals)

    hic.maskChromosomes(['c'])

    print(hic.matrix)


def test_create_from_cool():
    hic_ma = hm.hiCMatrix(ROOT + 'one_interaction_4chr.cool')
    nt.assert_equal(sorted(hic_ma.matrix.indices), [0, 3])
    nt.assert_equal(sorted(hic_ma.matrix.data), [1, 1])
    nt.assert_equal(sorted(hic_ma.nan_bins)[:5], [1, 2, 4, 5, 6])
    hic_ma = hm.hiCMatrix(ROOT + 'one_interaction_diag_4chr.cool')
    nt.assert_equal(sorted(hic_ma.matrix.indices), [0])
    nt.assert_equal(sorted(hic_ma.matrix.data), [1])
    nt.assert_equal(sorted(hic_ma.nan_bins)[:5], [1, 2, 3, 4, 5])
    hic_ma.maskBins(hic_ma.nan_bins)
    assert hic_ma.matrix.shape == (1, 1)
    assert hic_ma.getBinSize() == 50000


def test_load_cool_matrix_only():
    hic_cool = hm.hiCMatrix(ROOT + 'Li_et_al_2015.cool', pUpperTriangleOnly=True)

    hic_cool_matrix_only = hm.hiCMatrix(ROOT + 'Li_et_al_2015.cool', pUpperTriangleOnly=True, pLoadMatrixOnly=True)
    instances = hic_cool_matrix_only.matrix[0]
    features = hic_cool_matrix_only.matrix[1]
    data = hic_cool_matrix_only.matrix[2]

    instances_cool, features_cool = hic_cool.matrix.nonzero()
    nt.assert_equal(hic_cool.matrix.data, data)
    nt.assert_equal(instances_cool, instances)
    nt.assert_equal(features_cool, features)
